/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: af_hci.c,v 1.79 2002/12/04 17:29:07 kds Exp $

   AF_AFFIX - HCI Protocol Address family for socket interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <linux/poll.h>
#include <linux/kmod.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <linux/utsname.h>

#include <linux/in.h>
#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/rtnetlink.h>

/* Local Includes */
#define FILEBIT	DBAFHCI

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/hci_mgr.h>


typedef struct hci_sock	hci_sock;

struct hci_sock{
	struct list_head	q;

	struct sock	*sk;

	BD_ADDR		bda;
	int		hdrlen;
	hci_con		*con;

	int		mode;
	hci_struct	*hci;
	hci_listener_t	*evl;		/* events are here */
	hci_lock_t	*lock;		/* waiting here */

	unsigned long	flags;

};

static inline hci_sock *hpf_get(struct sock *sk)
{
	return (hci_sock*)sk->prot;
}

static inline  void hpf_lock(hci_sock *hsk, __u16 opcode)
{
	hsk->lock = hci_lock(hsk->hci, opcode);
}

static inline void hpf_unlock(hci_sock *hsk)
{
	if (hsk->lock) {
		hci_unlock(hsk->lock);
		hsk->lock = NULL;
	}
}


/* HPF driver variables */
btlist_head_t			hci_socks;
spinlock_t			hpf_spinlock = SPIN_LOCK_UNLOCKED;

extern btlist_head_t		listeners;
extern btlist_head_t		hcimgrs;


/*
 * debuging stuff
 */
static long sock_count = 0;


/***************************    Connection Management   ********************************/

struct sock *hpf_lookup_sock(hci_struct *hci, BD_ADDR *bda)
{
	hci_sock	*hsk;
	struct sock	*sk = NULL;
	
	btl_read_lock(&hci_socks);
	btl_for_each (hsk, hci_socks)
		if (hsk->sk->state == CON_LISTEN && 
				(bda_zero(&hsk->bda) || bda_equal(&hci->bda, &hsk->bda))) {
			sock_hold(hsk->sk);
			sk = hsk->sk;
			break;
		}
	btl_read_unlock(&hci_socks);

	return sk;
}
	
hci_sock *hpf_alloc(int priority)
{
	hci_sock	*hsk;

	hsk = (hci_sock*)kmalloc(sizeof(hci_sock), priority);
	if (hsk == NULL)
		return NULL;
	memset(hsk, 0, sizeof(hci_sock));

	return hsk;
}

void hpf_destruct(struct sock *sk)
{
	hci_sock	*hsk = hpf_get(sk);

	DBFENTER;
	
	kfree(hsk);
	sock_count--;	/* XXX */

	DBFEXIT;
}

void hpf_destroy(struct sock *sk)
{
	hci_sock	*hsk = hpf_get(sk);
	
	DBFENTER;

	btl_unlink(&hci_socks, hsk);

	if (hsk->hci)
		hci_put(hsk->hci);

	if (hsk->con) {
		hcc_put(hsk->con);
		hsk->con = NULL;
	}

	if (sk->pair) {
		sock_put(sk->pair);
		sk->pair = NULL;
	}

	if (hsk->mode & AFFIX_MODE_MASK)
		affix_ctl_mode &= ~(hsk->mode & AFFIX_MODE_MASK);

	if (hsk->mode & AFFIX_MODE_HCI)
		hci_remove_listener(hsk->evl);
	else if (hsk->mode & AFFIX_MODE_MGR)
		btl_unlink(&hcimgrs, hsk->evl);
	else {
		if (sk->protocol == BTPROTO_HCIACL || sk->protocol == BTPROTO_HCISCO) {
			// AFFIX_MODE_SCO
			struct sk_buff	*skb;

			DBPRT(DBCTRL, "queue len: %d\n", skb_queue_len(&sk->receive_queue));
			
			/* remove all packets and check if threre is a connection in the queue */
			while ((skb = skb_dequeue(&sk->receive_queue)) != NULL) {
				if (skb->sk != sk)	/* A pending connection */
					hpf_destroy(skb->sk);
				kfree_skb(skb);
			}
		}
	}

	hci_free_listener(hsk->evl);

	sock_put(sk);

	DBFEXIT;
}

struct sock *hpf_create(int family, int priority, int zero_it)
{
	struct sock	*sk;
	hci_sock	*hsk;

	DBFENTER;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	sk  = sk_alloc(family, priority, zero_it);
#else
	sk  = sk_alloc(family, priority, zero_it, NULL);
#endif
	if (sk == NULL)
		return NULL;

	hsk = hpf_alloc(priority);
	if (hsk == NULL) {
		sk_free(sk);
		return NULL;
	}

	hsk->sk = sk;
	sk->prot = (void*)hsk;

	hsk->evl = hci_alloc_listener(priority);
	if (hsk->evl == NULL) {
		kfree(hsk);
		sk_free(sk);
		return NULL;
	}
	hsk->evl->data = sk;

	//__btl_add_tail(&hci_socks, hsk);

	sock_count++;	/* XXX */

	DBFEXIT;
	return sk;
}

void hpf_init_data(struct socket *sock, struct sock *sk)
{
	DBFENTER;
	
	sock_init_data(sock, sk);
	sk->destruct = hpf_destruct;
	
	DBFEXIT;
}

int hpf_wait_for_state(struct sock *sk, __u8 state, int noblock)
{
	int	err = 0;
	long	timeo;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	timeo = sock_sndtimeo(sk, noblock);

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	while (sk->state != state) {
		if (!timeo) {
			err = -EAGAIN;
			break;
		}
		if ((err = sock_error(sk))) 
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		timeo = schedule_timeout(timeo);
//		schedule();
		set_current_state(TASK_INTERRUPTIBLE);
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
	return err;
}

int hpf_connect_req(struct sock *sk, struct sockaddr_affix *addr)
{
	int		err = 0;
	hci_struct	*hci;
	hci_sock	*hsk = hpf_get(sk);

	DBFENTER;

	if (!bda_zero(&addr->local)) {
		hci = hci_lookup_bda(&addr->local);	/* find device */
		if (hci == NULL)
			return -ENODEV;
	} else
		hci = NULL;

	hsk->bda = addr->bda;

	if (sk->protocol == BTPROTO_HCIACL) {
		err = lp_connect_req(hci, &addr->bda, &hsk->con);
		if (err) {
			if (hci)
				hci_put(hci);
			return err;
		}

		hsk->hdrlen = hsk->con->hdrlen;
		hsk->hci = hci;
		if (hcc_linkup(hsk->con))
			sk->state = CON_ESTABLISHED;

	} else if (sk->protocol == BTPROTO_HCISCO) {
		err = lp_add_sco(hci, &addr->bda, &hsk->con);
		if (err) {
			if (hci)
				hci_put(hci);
			return err;
		}

		hsk->hdrlen = hsk->con->hdrlen;
		hsk->hci = hci;
	}

	DBFEXIT;
	return err;
}

void hpf_shutdown(struct sock *sk, int reason)
{
	if (reason < 0 )
		sk->err = -reason;
	else if (reason == 0xEEEE)
		sk->err = ETIMEDOUT;
	else 
		sk->err = ECONNRESET;

        sk->shutdown |= SEND_SHUTDOWN;
	sk->state = CON_CLOSED;
	if (!sk->dead) {
                sk->dead = 1;
		sk->state_change(sk);
        }
}

/* 
   HPF protocol subsystem.  Client of HCI. Callback functions
*/

int hpf_data_ind(hci_con *con, struct sk_buff *skb)
{
	int		err = 0;
	hci_sock	*hsk, *prev = NULL;
	struct sk_buff	*cskb;
	
	DBFENTER;

	btl_read_lock(&hci_socks);
	btl_for_each (hsk, hci_socks) {
		if (hsk->con == con) {
			if (prev) {
				cskb = skb_clone(skb, GFP_ATOMIC);
				if (cskb == NULL)
					break;
				err = affix_sock_rx(prev->sk, cskb);
				if (err) {
					BTERROR("Unable to queue HCI packet\n");
					kfree_skb(cskb);
				}
			}
			prev = hsk;
		}
	}
	if (prev) {
		/* do not clone last */
		err = affix_sock_rx(prev->sk, skb);
		if (err) {
			BTERROR("Unable to queue HCI packet\n");
			kfree_skb(skb);
		}
	}
	btl_read_unlock(&hci_socks);

	if (prev == NULL)
		kfree_skb(skb);	/* not found any */

	DBFEXIT;
	return 0;
}


struct sock *hpf_dup(struct sock *osk)
{
	struct sock	*sk;

	sk = hpf_create(PF_AFFIX, GFP_ATOMIC, 1);
	if (sk == NULL)
		return NULL;

	hpf_init_data(NULL, sk);

	sk->pair = osk;

	sk->protocol = osk->protocol;
	sk->type     = osk->type;
	sk->destruct = osk->destruct;
	sk->sleep    = osk->sleep;
	sk->priority = osk->priority;
	//	sk->socket   = osk->socket;

	return sk;
}

int hpf_connect_cfm(hci_con *con, int status)
{
	int		err = 0;
	struct sock	*sk, *pair;
	hci_sock	*hsk, *phsk;
	struct sk_buff	*skb;
	int		type;
	
	DBFENTER;

	btl_write_lock(&hci_socks);
	btl_for_each (phsk, hci_socks) {
		pair = phsk->sk;
		if (pair->socket == NULL)
			continue;
		if (pair->dead)
			continue;
		if (pair->protocol == BTPROTO_HCIACL)
			type = HCI_LT_ACL;
		else if (pair->protocol == BTPROTO_HCISCO )
			type = HCI_LT_SCO;
		else
			continue;
		if (con->link_type != type)
			continue;
		if (pair->state == CON_LISTEN) {
			if (bda_zero(&phsk->bda) || bda_equal(&con->hci->bda, &phsk->bda)) {
				if (pair->ack_backlog < pair->max_ack_backlog && status == 0) {
					DBPRT(DBCTRL, "Server socket found\n");
					sk = hpf_dup(pair);
					if (sk == NULL) {
						err = -ENOMEM;
						break;
					}
					__btl_add_tail(&hci_socks, hpf_get(sk));
					sock_hold(pair);
				} else
					continue;
			} else
				continue;
		} else if (pair->state == CON_CONNECTING) {
			if (bda_equal(&con->bda, &phsk->bda)) {
				sk = pair;
				if (status) {
					hpf_shutdown(sk, status);
					continue;
				}
			} else
				continue;
		} else
			continue;

		hsk = hpf_get(sk);
		// set attributes for hci socket
		hsk->hdrlen = con->hdrlen;
		sk->state = CON_ESTABLISHED;
		// bda, con, hci ???
		hsk->bda = con->bda;
		if (hsk->hci == NULL) {
			hsk->hci = con->hci;
			hci_hold(con->hci);
		}
		if (hsk->con == NULL) {
			hsk->con = con;
			hcc_hold(con);
		}
		if (sk->pair){ /* incomming connection */
			sk->pair->ack_backlog++;
			skb = alloc_skb(0, GFP_ATOMIC); 
			skb_set_owner_r(skb, sk);
			DBPRT(DBCTRL, "queue len: %d\n", skb_queue_len(&sk->pair->receive_queue));
			skb_queue_tail(&sk->pair->receive_queue, skb);
			sk->pair->data_ready(sk->pair, skb->len);
		} else
			sk->state_change(sk);
	}
	btl_write_unlock(&hci_socks);

	DBFEXIT;
	return 0;
}

int hpf_disconnect_ind(hci_con *con)
{
	hci_sock	*hsk;

	DBFENTER;

	btl_read_lock(&hci_socks);
	btl_for_each (hsk, hci_socks) {
		if (hsk->con == con) {
			if (!hsk->sk->dead)
				hpf_shutdown(hsk->sk, 0);
		}
	}
	btl_read_unlock(&hci_socks);


	DBFEXIT;
	return 0;
}


/******************************  HCI PROTOCOL FAMILY SUBSYSTEM  **************************/

static inline void hpf_wait_on_queue(struct sock *sk)
{
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	add_wait_queue(sk->sleep, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	if (skb_peek(&sk->receive_queue) == NULL)
		schedule();

	set_current_state(TASK_RUNNING);
	remove_wait_queue(sk->sleep, &wait);

	DBFEXIT;
}

int hpf_release(struct socket *sock)
{
	struct sock	*sk = sock->sk;
	hci_sock	*hsk;
	int		dec = 1, err;

	DBFENTER;

	if (sk == NULL)
		return 0;

        sock->sk   = NULL;      
        sk->socket = NULL;      /* Not used, but we should do this. */
	
	hsk = hpf_get(sk);

	if (hsk->mode & AFFIX_MODE_KERNEL)
		dec = 0;

	if (hsk->mode & AFFIX_MODE_SUPER) {
		if (hsk->hci) {
			hci_struct	*hci = hsk->hci;
			hci->open_count--;
			if (!hci->open_count &&
					(hci->flags & HCI_FLAGS_RUNNING) 
					&& !(hci->flags & HCI_FLAGS_UP)) {
				rtnl_lock();
				err = dev_close(hci->dev);
				rtnl_unlock();
				hci->flags &= ~(HCI_FLAGS_RUNNING | HCI_FLAGS_UP);
			}
		}
	}

	sk->state = CON_CLOSED;
        sk->err = 0;
	sk->shutdown |= SEND_SHUTDOWN;
	sk->dead = 1;

	hpf_destroy(sk);

	if (dec)
		MOD_DEC_USE_COUNT;

	DBFEXIT;
	return 0;
}

int hpf_bind(struct socket *sock, struct sockaddr *addr, int addr_len)
{
	int		err = 0;
	hci_sock	*hsk = hpf_get(sock->sk);
	struct sockaddr_affix	*saddr = (void*)addr;

	DBFENTER;

	//err = l2ca_register_protocol(saddr->psm, &hpf_proto, sk);
	//if (!err)
	//	sk->sport = saddr->psm;

	hsk->bda = saddr->bda;
	
	DBFEXIT;
	return err;
}

/*
    Connect to remote device
    Wait until connection established
*/
int hpf_connect(struct socket *sock, struct sockaddr *addr, int alen, int flags)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*saddr = (void*)addr;
	int			err = 0;

	DBFENTER;

	/* deal with restarts */
	if (sk->state == CON_ESTABLISHED && sock->state == SS_CONNECTING) {
		sock->state = SS_CONNECTED;
		return 0;
	}
		
	if (sk->state == CON_CLOSED && sock->state == SS_CONNECTING) {
		sock->state = SS_UNCONNECTED;
		return -ECONNREFUSED;
	}
	
	if (sk->state == CON_ESTABLISHED)
		return -EISCONN;      /* No reconnect on a seqpacket socket */
	
	/* Move to connecting socket, start sending Connect Requests */
	sock->state = SS_CONNECTING;
	sk->state   = CON_CONNECTING;

	err = hpf_connect_req(sk, saddr);
	if (err < 0)
		goto exit;

	err = hpf_wait_for_state(sk, CON_ESTABLISHED, flags & O_NONBLOCK);
	if (err)
		goto exit;

	if (sk->state == CON_ESTABLISHED) {
		sock->state = SS_CONNECTED;
	}

 exit:
	DBFEXIT;
	return err;
}

int hpf_open_hci(struct sock *sk, struct hci_open *param)
{
	int			err = 0;
	hci_struct		*hci = NULL;
	hci_sock		*hsk = hpf_get(sk);

	DBFENTER;

	if (param->cmd & HCI_OPEN_KERNEL) {
		hsk->mode |= AFFIX_MODE_KERNEL;
		MOD_DEC_USE_COUNT;
	}
	if (param->cmd & HCI_OPEN_SUPER)
		hsk->mode |= AFFIX_MODE_SUPER;

	hsk->mode &= ~AFFIX_MODE_TYPE;
	hsk->mode |= AFFIX_MODE_HCI;	/* default */

	switch (param->cmd & ~HCI_OPEN_MASK) {
	case HCI_OPEN_NAME:
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (strncmp(hci->name, param->name, IFNAMSIZ) == 0 &&
				hcidev_present(hci->dev) && 
				(hci_running(hci) || (param->cmd & HCI_OPEN_SUPER))) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (hci == NULL)
			return -ENODEV;
		break;
	case HCI_OPEN_ID:
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (hci->hd == param->hd && hcidev_present(hci->dev) && 
				(hci_running(hci) || (param->cmd & HCI_OPEN_SUPER))) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (hci == NULL)
			return -ENODEV;
		break;
	case HCI_OPEN_EVENT:
		hci = NULL;
		break;
	case HCI_OPEN_MGR:
		hsk->mode &= ~AFFIX_MODE_TYPE;
		hsk->mode |= AFFIX_MODE_MGR;
		if (!btl_linked(hsk->evl))
			btl_add_tail(&hcimgrs, hsk->evl);
		break;
	}

	if (hci && (hsk->mode & AFFIX_MODE_SUPER)) {
		/* open net device */
		if (!(hci->flags & HCI_FLAGS_RUNNING)) {
			rtnl_lock();
			err = dev_open(hci->dev);
			rtnl_unlock();
			if (err) {
				hci_put(hci);
				return err;
			}
			hci->flags |= HCI_FLAGS_RUNNING;
		}
		hci->open_count++;
	}

	hsk->hci = hci;
	hsk->evl->hci = hci;

	DBFEXIT;
	return err;
}

void hpf_cmd_handler(hci_listener_t *evl, __u8 event, struct sk_buff *skb)
{
	struct sock			*sk = (void*)evl->data;
	hci_sock			*hsk = hpf_get(sk);
	DBFENTER;

	if (evl->err) {
		__hci_remove_listener(evl);
		hpf_unlock(hsk);
		return;
	}
	if (event == HCI_E_COMMAND_STATUS) {
		struct Command_Status_Event	*cse = (void*)skb->data;
		evl->event_mask &= ~COMMAND_STATUS_MASK;/* we need only one */
		if (evl->event_mask == 0 || cse->Status != 0)
			__hci_remove_listener(evl);
		hpf_unlock(hsk);
	} else if  (event == HCI_E_COMMAND_COMPLETE) {
		evl->event_mask &= ~COMMAND_COMPLETE_MASK;/* we need only one */
		if (evl->event_mask == 0)
			__hci_remove_listener(evl);
		hpf_unlock(hsk);
	}

	DBFEXIT;
}

int hpf_exec_cmd(struct sock *sk, struct hci_cmd_arg *param)
{
	int		err = 0;
	__u16		OpCode;
	__u8		Length;
	struct sk_buff	*skb;
	signed long	timeout = 0;
	hci_sock	*hsk = hpf_get(sk);
	hci_struct	*hci = hsk->hci;

	DBFENTER;

	if (hsk->flags & HCI_COMMAND_PENDING)
		goto recv;
	if (get_user(OpCode, &param->cmd->OpCode))
		return -EFAULT;
	if (get_user(Length, &param->cmd->Length))
		return -EFAULT;
	if (hsk->hci == NULL) /* it's a HCI event socket */
		return -ENODEV;

	/*
	  Lock HCI before registering to prevent receiving events
	  from the same command
	*/
	hpf_lock(hsk, OpCode);

	/* set some fields */
	hsk->flags = param->flags;
	if (!(hsk->flags & HCI_SKIP_STATUS))
		hsk->flags |= HCI_W4_STATUS;

	/* register listener for results */
	hsk->evl->event_mask = param->event_mask;
	hsk->evl->handler = hpf_cmd_handler;
	hsk->evl->opcode = OpCode;	/* !!! in BT order */

	if (param->event_mask != 0)
		hci_add_listener(hsk->evl);

	if (hsk->mode & AFFIX_MODE_SUPER)
		hci->cmd_count++;	// allow certainly to send command

	/* send a command */
	err = hci_send_cmd(hci, OpCode, Length, param->cmd->Data);
	if (err)
		goto exit;

	hsk->flags |= HCI_COMMAND_PENDING;

	timeout = HZ*5;	// 5 sec.

 recv:
 	param->status = 0;
	while ((hsk->flags & 0x03) && err == 0) {
		if ((skb = skb_dequeue(&hsk->evl->queue)) == NULL) {
			err = hci_wait_event(hsk->evl, 1, timeout);
			if (err != -ERESTARTSYS)
				hsk->flags &= ~HCI_COMMAND_PENDING;
			if (err == -ETIMEDOUT)
				err = -EIO;
			if (err)
				goto exit;
		} else {
			HCI_Event_Packet_Header		*hdr = (void*)skb->data;
			struct Command_Status_Event	*cse = (void*)hdr;

			__u8	EventCode = hdr->EventCode;
			__u8	Length = hdr->Length + HCI_EVENT_HDR_LEN;
			
			if (EventCode == HCI_E_COMMAND_COMPLETE) {
				hsk->flags &= ~HCI_W4_STATUS;
				/* results of the command */
				if (param->event && copy_to_user(param->event, hdr, Length))
					err = -EFAULT;
				param->status = 0;

			} else if (EventCode == HCI_E_COMMAND_STATUS) {
				hsk->flags &= ~HCI_W4_STATUS;
				param->status = cse->Status;
			}
			kfree_skb(skb);
		}
	}
	hsk->flags &= ~HCI_COMMAND_PENDING;
 exit:
	hpf_unlock(hsk);
	DBPRT(DBCTRL, "err: %d\n", err);
	DBFEXIT;
	return err;
}

int hpf_listen_event(struct sock *sk, __u64 *event_mask)
{
	int	err = 0;
	hci_sock		*hsk = hpf_get(sk);

	DBFENTER;

	if (*event_mask == 0) {
		hci_remove_listener(hsk->evl);
		hsk->evl->event_mask = 0;
	} else {
		hsk->evl->event_mask = *event_mask;
		hci_add_listener(hsk->evl);
	}

	DBFEXIT;
	return err;
}

int hpf_recv_event(struct sock *sk, struct hci_event_arg *param)
{
	HCI_Event_Packet_Header	*hdr;
	int			err = 0;
	struct sk_buff		*skb;
	int			Length;
	hci_sock		*hsk = hpf_get(sk);

	DBFENTER;

	while ((skb = skb_dequeue(&hsk->evl->queue)) == NULL) {
		err = hci_wait_event(hsk->evl, 1, 0);
		if (err == -ENODEV && hsk->evl->hci == NULL)
			continue;
		else if (err)
			return err;
	}

	hdr = (void*)skb->data;
	Length = hdr->Length + HCI_EVENT_HDR_LEN;
	
	if (copy_to_user(param->event, hdr, Length))
		err = -EFAULT;

	param->hd = DEV_HCI(skb->dev)->hd;

	kfree_skb(skb);

	DBFEXIT;
	return err;
}

int hpf_recv_msg(struct sock *sk, __u8 *msg)
{
	int		err = 0;
	struct sk_buff	*skb;
	hci_sock		*hsk = hpf_get(sk);

	DBFENTER;

	while ((skb = skb_dequeue(&hsk->evl->queue)) == NULL) {
		err = hci_wait_event(hsk->evl, 1, 0);
		if (err) 
			return err;
	}

	if (copy_to_user(msg, skb->data, skb->len))
		err = -EFAULT;

	kfree_skb(skb);

	DBFEXIT;
	return err;
}

/* ------------------- device specific stuff ------------------ */

static inline int hpf_set_audio(hci_struct *hci, struct affix_audio *audio)
{
	int 	err = 0;
	int	fd;

	DBFENTER;

	if (!hci)
		return -ENODEV;

	hci->audio_mode = audio->mode;
	if (audio->setting) {
		hci->audio_setting = audio->setting;
		fd = hci_open_id(hci->hd);
		if (fd < 0)
			return fd;
		err = HCI_WriteVoiceSetting(fd, hci->audio_setting);
		err = (err<0)?-errno:err;
		hci_close(fd);
		if (err)
			return err;
	}
	rtnl_lock();
	err = hci->dev->do_ioctl(hci->dev, (void*)hci, SIOCHCI_SET_AUDIO);
	rtnl_unlock();
	DBFEXIT;
	return err;
}

static inline int hpf_get_conn(hci_struct *hci, struct hci_conn_info *ci)
{
	hci_con		*con;

	DBFENTER;
	if (!hci)
		return -ENODEV;
	con = hcc_lookup_acl(hci, &ci->bda);
	if (!con)
		return -EINVAL;
	ci->handle = con->chandle;
	hcc_put(con);
	DBFEXIT;
	return 0;
}

int hpf_get_devs(int *devs)
{
	hci_struct	*hci;
	int		count = 0;

	DBFENTER;
	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs) {
		if (hci->dev->deadbeaf)
			continue;
		devs[count++] = hci->hd;
		if (count == HCI_MAX_DEVS)
			break;
	}
	btl_read_unlock(&hcidevs);
	DBFEXIT;
	return count;
}

int hpf_set_secmode(hci_struct *hci, int sec_mode);

int hpf_start_dev(hci_struct *hci)
{
	int	err, fd;
	__u8	hci_version, lmp_version;
	__u16	hci_revision, manfid, lmp_subversion;

	DBFENTER;

	if (hci->flags & HCI_FLAGS_UP)
		return 0;
			
	hci->pkt_type = HCI_PT_DM1 | HCI_PT_DH1 | HCI_PT_DM3 | HCI_PT_DH3 | HCI_PT_DM5 | HCI_PT_DH5 | HCI_PT_HV1;

	fd = _hci_open_id(hci->hd);
	if (fd < 0) {
		hci_put(hci);
		return -ENODEV;
	}

	err = HCI_ReadBDAddr(fd, &hci->bda);
	if (err)
		goto initerr;
	bdacpy(hci->dev->dev_addr, &hci->bda);

	err = HCI_ReadBufferSize(fd, &hci->acl_mtu, &hci->sco_mtu, &hci->acl_num, &hci->sco_num);
	if (err)
		goto initerr;
	hci->acl_count = hci->acl_num;
	hci->sco_count = hci->sco_num;

	err = HCI_ReadLocalVersionInformation(fd, &hci_version, &hci_revision, 
						&lmp_version, &manfid, &lmp_subversion);
	if (err)
		goto initerr;

	hci->manfid = manfid;

	err = HCI_ReadLocalSupportedFeatures(fd, &hci->lmp_features);
	if (err)
		goto initerr;

	/* some general settings */
	err = HCI_WriteClassOfDevice(fd, HCI_COD_INFORMATION | HCI_COD_NETWORKING | HCI_COD_COMPUTER | HCI_COD_HANDPC);
	if (err)
		goto initerr;
		
	/* set some parameters */
	down_read(&uts_sem);
	err = HCI_ChangeLocalName(fd, system_utsname.nodename);
	up_read(&uts_sem);
	if (err)
		goto initerr;

	err = hpf_set_secmode(hci, HCI_SECURITY_SERVICE);
	if (err)
		goto initerr;

	err = HCI_WriteScanEnable(fd, 0x00);	// not visible by default
	if (err)
		goto initerr;
		
	if (!(hci->lmp_features & HCI_LF_SWITCH)) {
		hci->flags |= HCI_ROLE_DENY_SWITCH;	// deny switch if no support
		hci->flags &= ~HCI_ROLE_BECOME_MASTER;
	}
#ifdef CONFIG_AFFIX_DTL1_FIX
	if (manfid == HCI_LMP_NOKIA) {
		hci->fixit = 1;
		/*  Removed when UART RI flow control was introduced ??? */
		hci->acl_count = hci->acl_num = 1;
	}
#endif
	hci_close(fd);
	DBFEXIT;
	return 0;
initerr:
	err = (err<0)?-errno:err;
	hci_close(fd);
	return err;
}

static inline int hpf_setget_attr(hci_struct *hci, int get, struct hci_dev_attr *da)
{
	DBFENTER;
	if (!hci) {
		da->name[IFNAMSIZ-1] = '\0';
		btl_read_lock(&hcidevs);
		btl_for_each (hci, hcidevs) {
			if (((da->hd && da->hd == hci->hd) || (!da->hd && strncmp(hci->name, da->name, IFNAMSIZ) == 0))
						 && !hci->dev->deadbeaf) {
				hci_hold(hci);
				break;
			}
		}
		btl_read_unlock(&hcidevs);
		if (!hci)
			return -ENODEV;
	} else
		hci_hold(hci);
	if (get) {
		strncpy(da->name, hci->name, IFNAMSIZ);
		da->name[IFNAMSIZ-1] = '\0';
		da->bda = hci->bda;
		da->hd = hci->hd;
		da->flags = hci->flags;
	}
	hci_put(hci);
	DBFEXIT;
	return 0;
}

static inline int hpf_setget_ctl(struct sock *sk, int get, int *mode)
{
	int 		err = 0;
	hci_sock	*hsk = hpf_get(sk);

	DBFENTER;
	if (get) {
		*mode = affix_ctl_mode;
		return 0;
	}
	if (affix_ctl_mode & *mode)
		return -EBUSY;

	affix_ctl_mode |= *mode;
	hsk->mode |= *mode;
	DBFEXIT;
	return err;

}

static inline int hpf_setget_flags(hci_struct *hci, int get, int *flags)
{
	int 	err = 0;

	DBFENTER;

	if (!hci)
		return -ENODEV;

	if (get) {
		*flags = hci->flags;
		return 0;
	}
	/* up/down */
	if (!(hci->flags & HCI_FLAGS_RUNNING))
		return -EINVAL;

	if ((hci->flags ^ *flags) & HCI_FLAGS_UP) {
		if (hci->flags & HCI_FLAGS_UP) {
			hci->flags &= ~HCI_FLAGS_UP;
			hcidev_state_change(hci->dev, HCIDEV_DOWN);
		} else {
			err = hpf_start_dev(hci);
			if (err)
				return err;
			hci->flags |= HCI_FLAGS_UP;
			hcidev_state_change(hci->dev, HCIDEV_UP);
		}
	}
	DBFEXIT;
	return err;
}

static inline int hpf_setget_pkttype(hci_struct *hci, int get, int *pkt_type)
{
	DBFENTER;
	if (!hci)
		return -ENODEV;
	if (get) {
		*pkt_type = hci->pkt_type;
		return 0;
	}
	hci->pkt_type = (*pkt_type) ? *pkt_type:(HCI_PT_DM1|HCI_PT_DH1|HCI_PT_DM3|HCI_PT_DH3|HCI_PT_DM5|HCI_PT_DH5|HCI_PT_HV1);
	DBFEXIT;
	return 0;
}

int hpf_set_secmode(hci_struct *hci, int sec_mode)
{
	int	err = 0, fd;

	DBFENTER;
	if (!hci)
		return -ENODEV;
	
	sec_mode &= HCI_FLAGS_SECURITY;
	if (!sec_mode)
		return -EINVAL;
		
	if ((hci->flags & HCI_FLAGS_SECURITY) ^ sec_mode) {
		hci->flags = (hci->flags & ~HCI_FLAGS_SECURITY) | sec_mode;
		
		if (!(hci->flags & HCI_FLAGS_UP))
			return 0;
			
		fd = hci_open_id(hci->hd);
		if (fd < 0) {
			hci_put(hci);
			return -ENODEV;
		}

		err = HCI_WriteAuthenticationEnable(fd, (hci->flags & HCI_SECURITY_AUTH) != 0);
		if (err) {
			err = (err<0)?-errno:err;
			hci_close(fd);
			return err;
		}
		err = HCI_WriteEncryptionMode(fd, (hci->flags & HCI_SECURITY_ENCRYPT) != 0);
		err = (err<0)?-errno:err;
		hci_close(fd);
	}
	DBFEXIT;
	return err;
}

static inline int hpf_set_role(hci_struct *hci, int role)
{
	DBFENTER;
	if (!hci)
		return -ENODEV;
	role &= HCI_FLAGS_ROLE;
	hci->flags = (hci->flags & ~HCI_FLAGS_ROLE) | role;
	if (!(hci->lmp_features & HCI_LF_SWITCH)) {
		hci->flags |= HCI_ROLE_DENY_SWITCH;	// deny switch if no support
		hci->flags &= ~HCI_ROLE_BECOME_MASTER;
	}
	DBFEXIT;
	return 0;
}

static inline int hpf_setup_uart(hci_struct *hci, struct open_uart *uart)
{
	int	err;

	DBFENTER;
	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs) {
		if (strncmp(hci->name, uart->dev, IFNAMSIZ) == 0 
				&& !hci->dev->deadbeaf && hcidev_present(hci->dev)) {
			hci_hold(hci);
			break;
		}
	}
	btl_read_unlock(&hcidevs);
	if (!hci)
		return -ENODEV;
	rtnl_lock();
	err = hci->dev->do_ioctl(hci->dev, (void*)uart, BTIOC_SETUP_UART);
	rtnl_unlock();
	hci_put(hci);
	DBFEXIT;
	return err;
}

static struct affix_pan_operations	affix_pan_ops;

int affix_set_pan(struct affix_pan_operations *ops)
{
	DBFENTER;
	
	if (ops == NULL)
		affix_pan_ops.owner = NULL;
	else
		affix_pan_ops = *ops;
	
	DBFEXIT;
	return 0;
}


int hpf_init_pan(struct init_pan *arg)
{
	int	err = -ENODEV;

	DBFENTER;

	if (!affix_pan_ops.owner) {// || !try_inc_mod_count(affix_pan_ops.owner)) { 
		request_module("affix_pan");
		if (!affix_pan_ops.owner)// || !try_inc_mod_count(affix_pan_ops.owner))
			return -ENODEV;
	}

	if (affix_pan_ops.ioctl)
		err = affix_pan_ops.ioctl(BTIOC_INIT_PAN, (unsigned long)arg);

	//if (affix_uart_ops.owner)
	//	__MOD_DEC_USE_COUNT(affix_uart_ops.owner);

	DBFEXIT;
	return err;
}



/*
    our system calls for user space programs
*/
int hpf_ioctl(struct socket *sock, unsigned int cmd,  unsigned long arg)
{
	struct sock	*sk = sock->sk;
	hci_sock	*hsk = hpf_get(sk);
	int		err = 0;

	DBFENTER;

	if (_IOC_TYPE(cmd) == BTIOC_MAGIC) {
		switch(cmd) {
			case BTIOC_OPEN_HCI:
				err = hpf_open_hci(sk, (struct hci_open*)arg);
				break;
			case BTIOC_EXEC:
				err = hpf_exec_cmd(sk, (struct hci_cmd_arg*)arg);
				break;
			case BTIOC_LISTEN:
				err = hpf_listen_event(sk, (__u64*)arg);
				break;
			case BTIOC_RECV_EVENT:
				err = hpf_recv_event(sk, (struct hci_event_arg*)arg);
				break;
			case BTIOC_RECV_MSG:
				err = hpf_recv_msg(sk, (__u8*)arg);
				break;
			case BTIOC_DBMGET:
				*(__u32*)arg = affix_dbmask;
				break;
			case BTIOC_DBMSET:
				affix_dbmask = *(__u32*)arg;
				break;
			case BTIOC_ADDPINCODE:
				err = affix_add_pin((struct PIN_Code *)arg);
				break;
			case BTIOC_REMOVEPINCODE:
				err = affix_remove_pin((BD_ADDR*)arg);
				break;
			case BTIOC_ADDLINKKEY:
				err = affix_add_key((struct link_key*)arg);
				break;
			case BTIOC_REMOVELINKKEY:
				err = affix_remove_key((BD_ADDR*)arg);
				break;
			case BTIOC_SETUP_UART:
				err = hpf_setup_uart(hsk->hci, (struct open_uart*)arg);
				break;
			case BTIOC_OPEN_UART:
				err = affix_open_uart((struct open_uart*)arg);
				break;
			case BTIOC_CLOSE_UART:
				err = affix_close_uart((struct open_uart*)arg);
				break;
			case BTIOC_GETDEVS:
				err = hpf_get_devs((int*)arg);
				break;
			case BTIOC_SET_CTL:
				err = hpf_setget_ctl(sk, 0, (int*)arg);
				break;
			case BTIOC_GET_CTL:
				err = hpf_setget_ctl(sk, 1, (int*)arg);
				break;
			case BTIOC_SET_AUDIO:
				err = hpf_set_audio(hsk->hci, (struct affix_audio*)arg);
				break;
			case BTIOC_GET_CONN:
				err = hpf_get_conn(hsk->hci, (struct hci_conn_info*)arg);
				break;
			case BTIOC_GET_ATTR:
				err = hpf_setget_attr(hsk->hci, 1, (struct hci_dev_attr*)arg);
				break;
			case BTIOC_SET_ATTR:
				err = hpf_setget_attr(hsk->hci, 0, (struct hci_dev_attr*)arg);
				break;
			case BTIOC_SET_FLAGS:
				err = hpf_setget_flags(hsk->hci, 0, (int*)arg);
				break;
			case BTIOC_GET_FLAGS:
				err = hpf_setget_flags(hsk->hci, 1, (int*)arg);
				break;
			case BTIOC_SET_PKTTYPE:
				err = hpf_setget_pkttype(hsk->hci, 0, (int*)arg);
				break;
			case BTIOC_GET_PKTTYPE:
				err = hpf_setget_pkttype(hsk->hci, 1, (int*)arg);
				break;
			case BTIOC_SET_SECMODE:
				err = hpf_set_secmode(hsk->hci, *(int*)arg);
				break;
			case BTIOC_SET_ROLE:
				err = hpf_set_role(hsk->hci, *(int*)arg);
				break;
				/*
				 * PAN stuff
				 */
			case BTIOC_INIT_PAN:
				err = hpf_init_pan((struct init_pan*)arg);
				break;

			default:
				return -ENOIOCTLCMD;
		}
	} else {
		/* deliver it to device driver */
		if (hsk->hci) {
			rtnl_lock();
			err = hsk->hci->dev->do_ioctl(hsk->hci->dev, (void*)arg, cmd);
			rtnl_unlock();
		}
	}
	DBFEXIT;
	return err;
}


/*
   register protocol and allow clients to establish connection
*/
int hpf_listen(struct socket *sock, int backlog)
{
	struct sock	*sk = sock->sk;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED)
		return -EINVAL;

	//if (sk->sport == 0)
	//	return -EOPNOTSUPP;	

	if (sk->state != CON_LISTEN) {
		sk->max_ack_backlog = backlog;
		sk->state = CON_LISTEN;
		sock->flags |= __SO_ACCEPTCON;
	}
	
	DBFEXIT;
	return 0;
}


/*
   Accept connection from remote device
   Wait until ready connection will become available
*/

int hpf_accept(struct socket *sock, struct socket *newsock, int flags)
{
	struct sock	*sk = sock->sk, *newsk;
	struct sk_buff	*skb;
	int		err = -EINVAL;

	DBFENTER;

	if (sock->state != SS_UNCONNECTED || !(sock->flags & __SO_ACCEPTCON))
		goto exit;
	
	while( (skb = skb_dequeue(&sk->receive_queue)) == NULL) {
		if (flags & O_NONBLOCK)
			return -EWOULDBLOCK;

		hpf_wait_on_queue(sk);
		if (signal_pending(current)) 
			return -ERESTARTSYS;
	}

	if (newsock->sk) {
		sk_free(newsock->sk);
		newsock->sk = NULL;
	}

	DBPRT(DBCTRL, "queue len: %d\n", skb_queue_len(&sk->receive_queue));
	
	newsk = skb->sk;
	sock_put(newsk->pair);		/* put pair */
	newsk->pair = NULL;
	sock_graft(newsk, newsock);
	//newsk->state = CON_ESTABLISHED;
	newsock->state = SS_CONNECTED;
	sk->ack_backlog--;

	kfree_skb(skb);
	MOD_INC_USE_COUNT;
	err = 0;
 exit:			
	DBFEXIT;
	return err;
}

int hpf_getname(struct socket *sock, struct sockaddr *uaddr, int *alen, int peer)
{
	struct sock		*sk = sock->sk;
	struct sockaddr_affix	*addr = (struct sockaddr_affix*)uaddr;
	hci_sock		*hsk = hpf_get(sk);

	DBFENTER;

	addr->family = AF_AFFIX;
	memcpy(&addr->bda, &hsk->bda, 6);

	*alen = sizeof(struct sockaddr_affix);

	DBFEXIT;
	return 0;

}

int hpf_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
	//hci_sock		*hsk = (void*)sock->sk;

	DBFENTER;

	DBFEXIT;
	return 0;
}

/*
    size - size of the iovec array, size of the message
*/
int hpf_sendmsg(struct socket *sock, struct msghdr *msg, int size, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	hci_sock		*hsk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;

	DBFENTER;

	if (msg->msg_flags & ~MSG_DONTWAIT)
		return -EINVAL;

	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;

	if (sk->shutdown & SHUTDOWN_MASK) {
		send_sig(SIGPIPE, current, 0);
		return -EPIPE;
	}

	if (sk->state != CON_ESTABLISHED)
		return -ENOTCONN;

	hsk = hpf_get(sk);

	size = btmin(hsk->con->mtu, size);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 4)
	skb = sock_alloc_send_skb(sk, size + hsk->hdrlen, 0,
				  msg->msg_flags & MSG_DONTWAIT, &err);
#else
	skb = sock_alloc_send_skb(sk, size + hsk->hdrlen,
				  msg->msg_flags & MSG_DONTWAIT, &err);
#endif
	if (skb == NULL)
		return err;

	skb_reserve(skb, hsk->hdrlen);
	
	err = memcpy_fromiovec(skb_put(skb, size), msg->msg_iov, size);
	if (err)
		goto exit;

	/* send skb */
	err = hci_send(hsk->con, skb);
	if (err)
		goto exit;

	err = size;
 exit:
	DBFEXIT;
	return err;
}

int hpf_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
{
	struct sock		*sk = sock->sk;
	int			err = -ENOTCONN;
	struct sk_buff		*skb;
	int			copied;

	DBFENTER;

	if (sock->flags & __SO_ACCEPTCON)
		return(-EINVAL);

	if (sock->state != SS_CONNECTED) 
		return -ENOTCONN;

	if (sk->state != CON_ESTABLISHED)  {
		if (sk->shutdown & SHUTDOWN_MASK) {
			//send_sig(SIGPIPE, current, 0);
			return sock_error(sk);
		}
		return -ENOTCONN;
	}

	skb = skb_recv_datagram(sk, (flags & ~MSG_DONTWAIT) | MSG_PEEK, flags & MSG_DONTWAIT, &err);
	if (skb == NULL)
		goto exit;

	DBPRT(DBRECV, "We received %d bytes\n", skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);

	skb->h.raw = skb->data;

	if (size < skb->len) {
		copied = size;
		if (sock->type == SOCK_SEQPACKET)
			msg->msg_flags |= MSG_TRUNC;
	} else
		copied = skb->len;
	skb_copy_datagram_iovec(skb, 0, msg->msg_iov, copied);
	if (!(flags & MSG_PEEK)) {
		if (msg->msg_flags & MSG_TRUNC)
			skb->len = 0;
		else
			skb_pull(skb, copied);
	}
	if (!skb->len) {
		skb_unlink(skb);
		kfree_skb(skb);
	}
	kfree_skb(skb);		// remove MSG_PEEK link
	err = copied;
 exit:
	DBFEXIT;
	return err;
}

unsigned int hpf_poll(struct file *filp, struct socket *sock, poll_table *wait)
{
	struct sock	*sk = sock->sk;
	hci_sock	*hsk = hpf_get(sk);
	unsigned int	mask = 0;

	DBFENTER;

	poll_wait(filp, &hsk->evl->sleep, wait);

	if (!skb_queue_empty(&hsk->evl->queue))
		mask |= POLLIN | POLLRDNORM;

	DBFEXIT;

	return mask;
}


/*
    HCI Protocol Family operations
*/

struct proto_ops hpf_hci_ops = {
	family:		PF_AFFIX,

	release:	hpf_release,
	bind:		sock_no_bind,//hpf_bind,
	connect:	sock_no_connect,//hpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		sock_no_accept,//hpf_accept,
	getname:	hpf_getname,
	poll:		hpf_poll,
	ioctl:		hpf_ioctl,
	listen:		sock_no_listen,//hpf_listen,
	shutdown:	sock_no_shutdown,
	setsockopt:	hpf_setsockopt,
	getsockopt:	sock_no_getsockopt,
	sendmsg:	sock_no_sendmsg,//hpf_sendmsg,
	recvmsg:	sock_no_recvmsg,//hpf_recvmsg,
	mmap:		sock_no_mmap
};

/*
 * ops for ACL/SCO sockets
 */
struct proto_ops hpf_data_ops = {
	family:		PF_AFFIX,

	release:	hpf_release,
	bind:		hpf_bind,
	connect:	hpf_connect,
	socketpair:	sock_no_socketpair,
	accept:		hpf_accept,
	getname:	hpf_getname,
	poll:		datagram_poll,
	ioctl:		hpf_ioctl,
	listen:		hpf_listen,
	shutdown:	sock_no_shutdown,
	setsockopt:	hpf_setsockopt,
	getsockopt:	sock_no_getsockopt,
	sendmsg:	hpf_sendmsg,
	recvmsg:	hpf_recvmsg,
	mmap:		sock_no_mmap
};


int __hpf_create(struct socket *sock, int protocol)
{
	struct sock		*sk;

	DBFENTER;

	sock->state = SS_UNCONNECTED;

	if (sock->type != SOCK_RAW)
		return -ESOCKTNOSUPPORT;

	sk = hpf_create(PF_AFFIX, GFP_KERNEL, 1);
	if (sk == NULL)
		return -ENOMEM;

	hpf_init_data(sock, sk);	/* generic initialization */
	sk->protocol = protocol;

	if (protocol == BTPROTO_HCIACL || protocol == BTPROTO_HCISCO)
		sock->ops = &hpf_data_ops;
	else
		sock->ops = &hpf_hci_ops;

	btl_add_tail(&hci_socks, hpf_get(sk));

	MOD_INC_USE_COUNT;

	DBFEXIT;
	return 0;
}


struct net_proto_family hpf_family_ops = {
	family:	PF_AFFIX,
	create:	__hpf_create
};



int hpf_proc_read(char *buf, char **start, off_t offset, int len)
{
	int count = 0;

	DBFENTER;

	count += sprintf(buf+count, "HCI sockets in use: %ld, list len: %d\n",
			sock_count, hci_socks.len);

	DBFEXIT;
	return count;
}


struct proc_dir_entry	*hpf_proc;

int init_hpf(void)
{
	int	err;
	DBFENTER;

	hpf_proc = create_proc_info_entry("hci_sock", 0, proc_affix, hpf_proc_read);
	if (hpf_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		return -EINVAL;
	}

	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCI);
	if (err)
		goto err1;

	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCIACL);
	if (err)
		goto err2;

	err = affix_sock_register(&hpf_family_ops, BTPROTO_HCISCO);
	if (err)
		goto err3;

	btl_head_init(&hci_socks);
	
	DBFEXIT;
	return 0;
err3:
	affix_sock_unregister(BTPROTO_HCIACL);
err2:
	affix_sock_unregister(BTPROTO_HCI);
err1:
	remove_proc_entry("hci_sock", proc_affix);
	return err;
}

void exit_hpf(void)
{
	DBFENTER;

	affix_sock_unregister(BTPROTO_HCISCO);
	affix_sock_unregister(BTPROTO_HCIACL);
	affix_sock_unregister(BTPROTO_HCI);
	remove_proc_entry("hci_sock", proc_affix);

	DBFEXIT;
}


static int	errno;

#include <affix/hci_cmds.h>	//XXX

/* Exporting stuff */
EXPORT_SYMBOL(hci_open_dev);
#if defined(CONFIG_AFFIX_PAN)
EXPORT_SYMBOL(HCI_SwitchRole);
EXPORT_SYMBOL(HCI_WriteScanEnable);
EXPORT_SYMBOL(HCI_ReadScanEnable);
EXPORT_SYMBOL(HCI_WritePageScanActivity);
EXPORT_SYMBOL(HCI_WriteInquiryScanActivity);
EXPORT_SYMBOL(HCI_WriteClassOfDevice);
EXPORT_SYMBOL(HCI_WriteLinkSupervisionTimeout);
EXPORT_SYMBOL(HCI_ReadBDAddr);
EXPORT_SYMBOL(HCI_ReadRSSI);
EXPORT_SYMBOL(HCI_GetLinkQuality);
EXPORT_SYMBOL(HCI_Read_Transmit_Power_Level);
EXPORT_SYMBOL(HCI_Inquiry);
EXPORT_SYMBOL(HCI_ChangeConnectionPacketType);
#endif

/*
 * 1. set mode mask for connecting sockets
 * 2. disconnect based on mode mask
 */

