/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: hci.c,v 1.106 2002/12/02 16:50:42 kds Exp $

   Host Controller Interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/notifier.h>
#include <linux/poll.h>
#include <linux/kmod.h>
#include <linux/major.h>
#include <linux/termios.h>

#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/in.h>
#include <linux/netdevice.h>
#include <linux/rtnetlink.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <net/pkt_sched.h>

#if defined(CONFIG_USB) || defined(CONFIG_USB_MODULE)
#include <linux/usb.h>
#endif


/* Local Includes */
#define FILEBIT	DBHCI

#undef CONFIG_AFFIX_OLDRX

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/hci_mgr.h>


struct proc_dir_entry	*proc_affix;

__u32	affix_dbmask = CONFIG_AFFIX_DBMASK;

btlist_head_t		hcidevs;
btlist_head_t		hcicons;
btlist_head_t		listeners;
btlist_head_t		hcimgrs;

struct notifier_block		*affix_chain=NULL;
struct affix_uart_operations	affix_uart_ops;


/***************************  CONNECTION MANAGEMENT ********************************/

void hcc_timer(unsigned long p)
{
	hci_con *con = (hci_con *)p;
	
	DBFENTER;

	switch (STATE(con)) {
		case CON_OPEN:
			if (atomic_read(&con->refcnt) == 1 ) {
				lp_disconnect_req(con, 0x13);
			} else {
				if (con->btdev) {
					if (con->btdev->state == BTDEV_AUTHPENDING) {
						con->btdev->state = 0;
						con->btdev->paired = 0;
						notifier_call_chain(&affix_chain, HCICON_AUTH_COMPLETE, con);
					}
				}
			}
			break;
		case CON_W4_CONRSP:
		case CON_W4_LCONREQ:
			ENTERSTATE(con, DEAD);
			lp_connect_cfm(con, HCI_ERR_HARDWARE_FAILURE);
			break;
		default:
			/* DEAD | CON_W4_DISCRSP  - just release it */
	}
	hcc_put(con);

	DBFEXIT;
}

hci_con *__hcc_lookup_acl(hci_struct *hci, BD_ADDR *bda)
{
	hci_con	*con;

	btl_for_each (con, hcicons) {
		if (STATE(con) != DEAD && STATE(con) != CON_W4_DISCRSP && con->link_type == HCI_LT_ACL &&
				(!hci || con->hci == hci) && !memcmp(&con->bda, bda, 6))
				return con;
	}
	return NULL;
}

hci_con *hcc_lookup_acl(hci_struct *hci, BD_ADDR *bda)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_acl(hci, bda);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}

hci_con *__hcc_lookup_sco(hci_struct *hci, BD_ADDR *bda)
{
	hci_con	*con;

	btl_for_each (con, hcicons) {
		if (STATE(con) != DEAD && STATE(con) != CON_W4_DISCRSP && con->link_type == HCI_LT_SCO &&
				con->hci == hci && !memcmp(&con->bda, bda, 6))
				return con;
	}
	return NULL;
}

hci_con *hcc_lookup_sco(hci_struct *hci, BD_ADDR *bda)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_sco(hci, bda);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}


hci_con *__hcc_lookup_chandle(hci_struct *hci, __u16 chandle)
{
	hci_con	*con;
	
	btl_for_each (con, hcicons) {
		if (STATE(con) != DEAD &&
				con->hci == hci && con->chandle == chandle)
				return con;
	}
	return NULL;
}

hci_con *hcc_lookup_chandle(hci_struct *hci, __u16 chandle)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_chandle(hci, chandle);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}

hci_con *__hcc_lookup_id(int id)
{
	hci_con	*con;

	btl_for_each (con, hcicons)
		if (STATE(con) != DEAD && con->id == id)
			return con;
	return NULL;
}

hci_con *hcc_lookup_id(int id)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_id(id);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}


void hcc_bind(hci_con *con, hci_struct *hci)
{
	hci_hold(hci);
	con->hci = hci;
	if (con->link_type == HCI_LT_ACL) {
		con->hdrlen = hci->hdrlen + HCI_ACL_HDR_LEN;
		con->mtu = hci->acl_mtu;
	} else {
		con->hdrlen = hci->hdrlen + HCI_SCO_HDR_LEN;
		con->mtu = hci->sco_mtu;
	}
}
	
hci_con *hcc_alloc(void)
{
	hci_con	*con = NULL;
	static int	hcc_desc = 1;
	
	DBFENTER;

	con = kmalloc(sizeof(hci_con), GFP_ATOMIC);
	if (con == NULL)
		return NULL;
	memset(con, 0, sizeof(hci_con));

	con->id = hcc_desc++;
	atomic_set(&con->refcnt, 1);

	/* by default it's an ACL connection */
	con->retry = 1;
	con->link_type = HCI_LT_ACL;

	skb_queue_head_init(&con->tx_queue);

	init_timer(&con->timer);
	con->timer.data = (unsigned long)con;
	con->timer.function = hcc_timer;

	DBFEXIT;
	return con;
}

/*
    Create new connection object. In fact should work atomicaly
    If the object exists it returns existing
*/
hci_con *hcc_create(hci_struct *hci, BD_ADDR *bda)
{
	hci_con	*con;

	DBFENTER;

	btl_write_lock(&hcicons);
	con = __hcc_lookup_acl(hci, bda);
	if (!con) {
		con = hcc_alloc();
		if (!con)
			goto exit;
		con->btdev = btdev_create(bda);
		if (!con->btdev) {
			kfree(con);
			con = NULL;
			goto exit;
		}
		memcpy(&con->bda, bda, 6);
		if (hci)
			hcc_bind(con, hci);
		ENTERSTATE(con, CON_CLOSED);	/* SET initial state */
		__btl_add_tail(&hcicons, con);
	} else
		hcc_hold(con);	/* inc refcnt */
exit:
	btl_write_unlock(&hcicons);

	DBFEXIT;
	return con;
}

hci_con *hcc_create_sco(hci_struct *hci, BD_ADDR *bda)
{
	hci_con		*con;

	DBFENTER;

	btl_write_lock(&hcicons);
	con = hcc_alloc();
	if (con) {
		con->link_type = HCI_LT_SCO;
		con->btdev = btdev_create(bda);
		if (!con->btdev) {
			kfree(con);
			con = NULL;
			goto exit;
		}
		memcpy(&con->bda, bda, 6);
		if (hci)
			hcc_bind(con, hci);
		ENTERSTATE(con, CON_CLOSED);	/* SET initial state */
		__btl_add_tail(&hcicons, con);
	}
exit:
	btl_write_unlock(&hcicons);

	DBFEXIT;
	return con;
}


void __hcc_destroy(hci_con *con)
{
	DBFENTER;

	if (atomic_read(&con->refcnt) == 0) {
		__btl_unlink(&hcicons, con);
		skb_queue_purge(&con->tx_queue);
		if (con->hci)
			hci_put(con->hci);
		if (con->btdev)
			btdev_put(con->btdev);
		kfree(con);
	}

	DBFEXIT;
}

void hcc_destroy(hci_con *con)
{
	btl_write_lock(&hcicons);
	__hcc_destroy(con);
	btl_write_unlock(&hcicons);
}

void hcc_put(hci_con *con)
{
	DBFENTER;

	if (con->state == CON_OPEN && atomic_read(&con->refcnt) == 1) {
		if (con->link_type == HCI_LT_ACL)
			hcc_start_timer(con, HCC_DEFER_DISC_TIMEOUT);
		else
			lp_disconnect_req(con, 0x13);
	}
	
	if (atomic_dec_and_test(&con->refcnt))
		hcc_destroy(con);

	DBFEXIT;
}

static inline void __hcc_put(hci_con *con)
{
	DBFENTER;

	if (atomic_dec_and_test(&con->refcnt))
		__hcc_destroy(con);

	DBFEXIT;
}


/*************************   ---- HCI ----    *******************************/

/*
 * destruct HCI object
 */
void hci_destructor(struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);

	DBFENTER;
	hci->dev = NULL;
	btl_unlink(&hcidevs, hci);
	/* tx cleanup */
	skb_queue_purge(&hci->cmd_queue);
	/* rx cleanup */
	tasklet_kill(&hci->rx_task);
	skb_queue_purge(&hci->rx_queue);
	DBFEXIT;
}

/*
 * create HCI object
 */
hci_struct *hci_create(struct net_device *dev)
{
	static int	hci_desc = 1;
	hci_struct	*hci = DEV_HCI(dev);

	DBFENTER;

	hci->hd = hci_desc++;
	strncpy(hci->name, dev->name, IFNAMSIZ);
	hci->dev = dev;
	hci_hold(hci);
	hci->hdrlen = dev->hard_header_len?dev->hard_header_len:1;

	/* rx init */
	tasklet_init(&hci->rx_task, hci_rx_task, (unsigned long) hci);
	tasklet_disable(&hci->rx_task);
	skb_queue_head_init(&hci->rx_queue);

	/* tx init */
	skb_queue_head_init(&hci->cmd_queue);
	btl_head_init(&hci->hci_locks);

	dev->destructor = hci_destructor;
	hci->flags = 0;

	btl_add_tail(&hcidevs, hci);	/* add to queue */
	
	DBFEXIT;
	return hci;
}

/*
  search for a hci devices
*/
hci_struct *__hci_lookup_bda(BD_ADDR *bda)
{
	hci_struct	*hci;

	DBFENTER;

	btl_for_each (hci, hcidevs)
		if (hci_running(hci) && memcmp(&hci->bda, bda, 6) == 0)
			return hci;

	DBFEXIT;
	return NULL;
}

hci_struct *hci_lookup_bda(BD_ADDR *bda)
{
	hci_struct	*hci;

	DBFENTER;

	btl_read_lock(&hcidevs);
	hci = __hci_lookup_bda(bda);
	if (hci)
		hci_hold(hci);
	btl_read_unlock(&hcidevs);

	DBFEXIT;
	return hci;
}

hci_struct *__hci_lookup(int hd)
{
	hci_struct	*hci;

	DBFENTER;

	btl_for_each (hci, hcidevs)
		if (hci_running(hci) && hci->hd == hd)
			return hci;

	DBFEXIT;
	return NULL;
}

hci_struct *hci_lookup(int hd)
{
	hci_struct	*hci;

	DBFENTER;

	btl_read_lock(&hcidevs);
	hci = __hci_lookup(hd);
	if (hci)
		hci_hold(hci);
	btl_read_unlock(&hcidevs);

	DBFEXIT;
	return hci;
}


/*
  search for any hci devices
*/
hci_struct *hci_select(void)
{
	hci_struct	*hci;

	DBFENTER;

	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs)
		if (hci_running(hci)) {
			hci_hold(hci);
			break;
		}
	btl_read_unlock(&hcidevs);

	DBFEXIT;
	return hci;
}


/* *******************       TRANSMITING SUBSYSTEM      *******************************/
int hci_send_cmd(hci_struct *hci, __u16 opcode, __u8 len, void *params)
{
	HCI_Command_Packet_Header	*hdr;
	hci_skb_cb	*cb;
	struct sk_buff	*skb;
	int		err;

	DBFENTER;

	if (!hci)
		return -ENODEV;

	skb = hci_alloc_cmd_skb(hci, HCI_CMD_HDR_LEN + len);
	if (!skb)
		return -ENOMEM;

	cb = (hci_skb_cb*)skb->cb;

	hdr = (HCI_Command_Packet_Header*)skb->data;
	hdr->OpCode = opcode;	/* BT byte order */
	hdr->Length = len;

	if (len != 0)
		if (copy_from_user(hdr->Data, params, len))
			return -EFAULT;

	cb->type = HCI_COMMAND;
	get_fast_time(&skb->stamp);
	skb->dev = hci->dev;

	DBPRT(DBCTRL, "CMD packet sent\n");
	DBPARSEHCI(DBCTRL, skb->data, skb->len, HCI_COMMAND, TO_HOSTCTRL);
	DBPRTPKT(DBCTRL, skb->data, skb->len);

	err = dev_queue_xmit(skb);

	DBFEXIT;
	return err;
}

/* 
     Send data from connection queue
*/
int hci_send_data(hci_con *con, struct sk_buff *skb)
{
	int		err;
	hci_struct	*hci = con->hci;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;

	DBFENTER;
	
	cb->type = HCI_ACL;
	cb->con = con;
	cb->pb = HCI_PB_FIRST;

	get_fast_time(&skb->stamp);
	skb->dev = hci->dev;

	DBPRT(DBXMIT, "ACL packet sent, length: %d\n", skb->len);
	DBPARSEHCI(DBXMIT, skb->data, skb->len, HCI_ACL, TO_HOSTCTRL);
	DBPRTPKT(DBXMIT, skb->data, skb->len);

	err = dev_queue_xmit(skb);

	DBFEXIT;
	return err;
}

int hci_send_voice(hci_con *con, struct sk_buff *skb)
{
	HCI_SCO_Packet_Header	*hdr;
	hci_struct	*hci = con->hci;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;
	__u16		len = skb->len;
	int		err;

	DBFENTER;

	hdr = (HCI_SCO_Packet_Header*)skb_push(skb, HCI_SCO_HDR_LEN);
	hdr->Connection_Handle = __htob16(con->chandle);
	hdr->Length = len;

	cb->type = HCI_SCO;
	cb->con = con;

	get_fast_time(&skb->stamp);
	skb->dev = hci->dev;

	err = dev_queue_xmit(skb);

	DBFEXIT;
	return err;
}

int hci_send(hci_con *con, struct sk_buff *skb)
{
	DBFENTER;

	/* check state of the channel */
	if (STATE(con) != CON_OPEN) {
		kfree_skb(skb);
		return -ENOTCONN;
	}

	if (con->link_type == HCI_LT_ACL)
		hci_send_data(con, skb);
	else
		hci_send_voice(con, skb);

	DBFEXIT;
	return 0;
}

/*
  Manager Subsystem
*/

int hci_deliver_msg(struct sk_buff *skb)
{
	hci_listener_t		*evl;
	struct sk_buff		*s;

	DBFENTER;

	btl_read_lock(&hcimgrs);

	btl_for_each (evl, hcimgrs) {
		s = skb_clone(skb, GFP_ATOMIC);
		if (s == NULL) {
			btl_read_unlock(&hcimgrs);
			return -1;
		}

		skb_queue_tail(&evl->queue, s);

		if (evl->handler)
			evl->handler(evl, event, s);

		if (waitqueue_active(&evl->sleep))
			wake_up_interruptible(&evl->sleep);
	}

	btl_read_unlock(&hcimgrs);

	kfree_skb(skb);

	DBFEXIT;
	return 0;
}

/*
    **************       HCI EVENT SUBSYSTEM            *****************
*/

/* event listeners */

hci_listener_t *hci_alloc_listener(int priority)
{
	hci_listener_t	*evl;

	evl = kmalloc(sizeof(hci_listener_t), priority);
	if (evl == NULL)
		return NULL;
	memset(evl, 0, sizeof(hci_listener_t));

	init_waitqueue_head(&evl->sleep);
	skb_queue_head_init(&evl->queue);

	return evl;
}

int hci_wait_event(hci_listener_t *evl, int in, signed long timeout)
{
	int	err = 0;
	DECLARE_WAITQUEUE(wait, current);

	DBFENTER;

	add_wait_queue(&evl->sleep, &wait);
	set_current_state(in?TASK_INTERRUPTIBLE:TASK_UNINTERRUPTIBLE);
	while (skb_peek(&evl->queue) == NULL) {
		if ((err = listener_error(evl)))
			break;
		if (signal_pending(current)) {
			err = -ERESTARTSYS;
			break;
		}
		if (timeout) {
			timeout = schedule_timeout(timeout);
			if (timeout == 0) {
				err = -ETIMEDOUT;
				break;
			}
		} else 
			schedule();
	
		set_current_state(in?TASK_INTERRUPTIBLE:TASK_UNINTERRUPTIBLE);
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&evl->sleep, &wait);

	DBFEXIT;
	return err;
}

int hci_deliver_event(hci_struct *hci, struct sk_buff *skb)
{
	struct Command_Complete_Event	*cce = (void*)skb->data;
	__u8				event = cce->hdr.EventCode;
	hci_listener_t			*evl, *next;
	struct sk_buff			*s;
	int	err = 0;

	DBFENTER;

	btl_write_lock(&listeners);
	btl_for_each_next (evl, listeners, next) {
		if ((evl->event_mask == ALL_EVENTS_MASK ||
					test_bit64(event-1, &evl->event_mask)) && 
		    (evl->hci == NULL || evl->hci == hci) ) {
			if (event == HCI_E_COMMAND_COMPLETE) {
				if (cce->Command_Opcode != evl->opcode)
					continue;
			} else if (event == HCI_E_COMMAND_STATUS) {
				struct Command_Status_Event	*h = (void*)skb->data;
				if (h->Command_Opcode != evl->opcode)
					continue;
			}
			DBPRT(DBCTRL, "Event queued\n");
			s = skb_clone(skb, GFP_ATOMIC);
			if (s == NULL) {
				err = -ENOMEM;
				break;
			}

			skb_queue_tail(&evl->queue, s);

			if (evl->handler)
				evl->handler(evl, event, s);

			if (waitqueue_active(&evl->sleep))
				wake_up_interruptible(&evl->sleep);
		}
	}
	btl_write_unlock(&listeners);

	DBFEXIT;
	return 0;
}

/*
    Check Number of Completed Packets event 
*/
int event_number_completed_packets(hci_struct *hci, struct sk_buff *skb)
{
	struct Number_Of_Completed_Packets	*event = (void*)skb->data;
	__u8		num = event->Number_Of_Handles;
	int		i;
	__u16		pnum;
	hci_con		*con;

	DBFENTER;
	for (i = 0; i < num; i++) {
		con = hcc_lookup_chandle(hci, HCI_HANDLE(event->Result[i].Connection_Handle));
		if (con) {
			pnum = __btoh16(event->Result[i].HC_Number_Of_Completed_Packets);
			con->prior -= pnum;
			if (con->link_type == HCI_LT_ACL)
				hci->acl_count += pnum;
			else
				hci->sco_count += pnum;
			hcc_put(con);
			DBPRT(DBCTRL, "Completed %d packets\n", pnum);
		} else {
			DBPRT(DBCTRL, "Connection does not exist\n");
		}

		DBPRT(DBCTRL, "FREE SPACE: acl - %d, sco - %d\n", hci->acl_count, hci->sco_count);
	}	
	DBFEXIT;
	return 0;
}

int event_flush_occurred(hci_struct *hci, struct sk_buff *skb)
{
	struct Flush_Occured_Event	*event = (void*)skb->data;
	hci_con		*con;
		
	DBFENTER;
	
	con = hcc_lookup_chandle(hci, HCI_HANDLE(event->Connection_Handle));
	if (con == NULL) {
		DBPRT(DBCTRL, "unknown connection !!!\n");
		return 0;
	}
	if (con->link_type == HCI_LT_ACL)
		hci->acl_count += con->prior;
	else
		hci->sco_count += con->prior;

	con->prior = 0;		/* we can send new */
	hcc_put(con);
	DBFEXIT;
	return 0;
}

int event_connection_request(hci_struct *hci, struct sk_buff *skb)
{
	struct Connection_Request_Event	*event = (void*)skb->data;
	hci_con		*con;
	int		ret = 0;

	DBFENTER;
	
	if (event->Link_Type == HCI_LT_ACL)
		con = hcc_create(hci, &event->bda);
	else
		con = hcc_create_sco(hci, &event->bda);

	if (!con) {
		return -ENOMEM;
	}

	/* expire pairing  */
	con->btdev->paired = 0;

	ENTERSTATE(con, CON_W4_LCONRSP);
	/* start timer to prevent stale connections - accept timeout */
	hcc_start_timer(con, HCC_CONN_TIMEOUT);
	hcc_put(con);
	
	DBFEXIT;
	return ret;
}

int event_connection_complete(hci_struct *hci, struct sk_buff *skb)
{
	struct Connection_Complete_Event	*evt = (void*)skb->data;
	hci_con		*con;
	int		ret = 0;

	DBFENTER;

	DBPRT(DBRECV, "Connection Complete, Status: 0x%02x, handle: %d, type: %s!!!\n",
			evt->Status, evt->Connection_Handle, (evt->Link_Type == HCI_LT_ACL)?"ACL":"SCO");
	if (evt->Link_Type == HCI_LT_SCO) {
		con = hcc_lookup_sco(hci, &evt->bda);
	} else {
		con = hcc_lookup_acl(hci, &evt->bda);
	}
	if (con == NULL) {
		DBPRT(DBRECV, "Unknown connection!!!\n");
		return -ENOTCONN;
	}
	/*
	 * check the state ??? CON_W4_CONRSP
	 */
	hcc_stop_timer(con);
	if (evt->Status) {
		ENTERSTATE(con, DEAD);
#ifdef CONFIG_AFFIX_L2CAP
		if (con->link_type == HCI_LT_ACL)
			lp_connect_cfm(con, evt->Status);
#endif
		hpf_connect_cfm(con, evt->Status);
		goto exit;
	}

	con->chandle = HCI_HANDLE(evt->Connection_Handle);
	//con->link_type = evt->Link_Type;
	con->encryp_mode = evt->Encryption_Mode;

	ENTERSTATE(con, CON_OPEN);

#ifdef CONFIG_AFFIX_L2CAP
	if (con->link_type == HCI_LT_ACL)
		lp_connect_cfm(con, 0);
#endif
	hpf_connect_cfm(con, 0);

exit:
	hcc_put(con);
	DBFEXIT;
	return ret;
}

int event_disconnection_complete(hci_struct *hci, struct sk_buff *skb)
{
	hci_con		*con;
	struct Disconnection_Complete_Event	*evt = (void*)skb->data;
	int		ret = 0;

	DBFENTER;

	con = hcc_lookup_chandle(hci, HCI_HANDLE(evt->Connection_Handle));
	if (con == NULL) { 
		DBPRT(DBRECV, "Unknown connection!!!\n");
		return -ENOTCONN;
	}
	hcc_stop_timer(con);
	ENTERSTATE(con, DEAD);
	if (con->link_type == HCI_LT_ACL)
		hci->acl_count += con->prior;
	else
		hci->sco_count += con->prior;

#ifdef CONFIG_AFFIX_L2CAP	
	lp_disconnect_ind(con);
#endif
	hpf_disconnect_ind(con);

	hcc_put(con);
	DBFEXIT;
	return ret;
}

int event_packet_type_changed(hci_struct *hci, struct sk_buff *skb)
{
	struct Connection_Packet_Type_Changed_Event	*event = (void*)skb->data;
	hci_con		*con;
		
	DBFENTER;
	
	con = hcc_lookup_chandle(hci, HCI_HANDLE(event->Connection_Handle));
	if (con == NULL)
		return 0;
		
	if (event->Status)
		goto exit;
	
	con->pkt_type = event->Packet_Type;
	DBPRT(DBCTRL, "Packet type: %x\n", con->pkt_type);
exit:
	hcc_put(con);
	DBFEXIT;
	return 0;
}

static inline int event_command_complete(hci_struct *hci, struct sk_buff *skb)
{
	struct Command_Complete_Event	*c = (void*)skb->data;
	
	hci->cmd_count = c->Num_HCI_Command_Packets;
	return 0;
}

int event_command_status(hci_struct *hci, struct sk_buff *skb)
{
	struct Command_Status_Event	*cs = (void*)skb->data;
	__u16	opcode;

	DBFENTER;
	
	hci->cmd_count = cs->Num_HCI_Command_Packets;

	if (cs->Status == 0)
		return 0;

	/* most of the *non-command_complete* commands return status on error */
	opcode = __btoh16(cs->Command_Opcode);
	switch (opcode) {
		case HCI_C_CREATE_CONNECTION: {
			hci_con		*con;

			con = hcc_lookup_acl(hci, &hci->target_dev.bda);
			if (con == NULL) { 
				DBPRT(DBRECV, "Unknown connection!!!");
				return -ENODEV;
			}
			ENTERSTATE(con, DEAD);
#ifdef CONFIG_AFFIX_L2CAP
			lp_connect_cfm(con, cs->Status);
#endif
			hpf_connect_cfm(con, cs->Status);
			hcc_put(con);
		      }
		break;

		default:
	}

	DBFEXIT;
	return 0;
}

int event_link_key_notification(hci_struct *hci, struct sk_buff *skb)
{
	struct Link_Key_Notification_Event	*evt = (void*)skb->data;
	btdev_struct	*btdev;

	DBFENTER;

	btdev = btdev_create(&evt->bda);
	if (btdev == NULL)
		return -ENOMEM;
	btdev->Key_Type = evt->Key_Type;
	memcpy(btdev->Link_Key, evt->Link_Key, 16);
	btdev->flags |= NBT_KEY;
	btdev_put(btdev);
	
	DBFEXIT;
	return 0;
}

int event_inquiry_result(hci_struct *hci, struct sk_buff *skb)
{
	struct Inquiry_Result_Event	*evt = (void*)skb->data;
	int		i;
	btdev_struct	*btdev;

	DBFENTER;

	for (i = 0; i < evt->Num_Responses; i++) {
		INQUIRY_ITEM	*item = &evt->Results[i];
		btdev = btdev_create(&item->bda);
		if (btdev == NULL)
			return -ENOMEM;
		/* set inquiry results */
		btdev->PS_Repetition_Mode = item->PS_Repetition_Mode;
		btdev->PS_Period_Mode = item->PS_Period_Mode;
		btdev->PS_Mode = item->PS_Mode;
		btdev->Class_of_Device = __btoh24(item->Class_of_Device);
		btdev->Clock_Offset = __btoh16(item->Clock_Offset);
		btdev->flags |= NBT_INQUIRY;
		btdev_put(btdev);
	}

	DBFEXIT;
	return 0;
}

int event_authentication_complete(hci_struct *hci, struct sk_buff *skb)
{
	struct Authentication_Complete_Event	*evt = (void*)skb->data;
	hci_con			*con;

	DBFENTER;

	con = hcc_lookup_chandle(hci, evt->Connection_Handle);
	if (!con) {
		return 0;
	}
	if (!con->btdev) {
		hcc_put(con);
		return -ENOTCONN;
	}
	con->btdev->state = 0;
	if (evt->Status == 0)
		con->btdev->paired = 1;
	else
		con->btdev->paired = 0;

	notifier_call_chain(&affix_chain, HCICON_AUTH_COMPLETE, con);
	hcc_put(con);

	DBFEXIT;
	return 0;
}


/*
 * events goes from BH only
 */
int hci_receive_event(hci_struct *hci, struct sk_buff *skb)
{
	HCI_Event_Packet_Header	*hdr = (HCI_Event_Packet_Header*)skb->data;
	__u8		event = hdr->EventCode;
	int		err = 0;
	int		flag = 0;

	DBFENTER;

	if (skb->len < HCI_EVENT_HDR_LEN && skb->len < (hdr->Length + HCI_EVENT_HDR_LEN)) {
		BTDEBUG("EVENT packet too small, drop it...\n");
		kfree_skb(skb);
		return -EINVAL;
	}

	DBPRT(DBCTRL, "HCI event received: 0x%02X\n", event);
	DBPARSEHCI(DBCTRL, skb->data, skb->len, HCI_EVENT, FROM_HOSTCTRL);	
	DBPRTPKT(DBCTRL, skb->data, skb->len);

	/* process events */
	switch (event) {
		case HCI_E_COMMAND_COMPLETE:
			event_command_complete(hci, skb);
			break;
		case HCI_E_COMMAND_STATUS:
			event_command_status(hci, skb);
			break;
		case HCI_E_NUMBER_COMPLETED_PACKETS:
			event_number_completed_packets(hci, skb);
			flag = 1;
			break;
		case HCI_E_FLUSH_OCCURRED:
			event_flush_occurred(hci, skb);
			flag = 1;
			break;
		case HCI_E_CONNECTION_REQUEST:
			event_connection_request(hci, skb);
			break;
		case HCI_E_CONNECTION_COMPLETE:
			event_connection_complete(hci, skb);
			break;
		case HCI_E_DISCONNECTION_COMPLETE:
			event_disconnection_complete(hci, skb);
			break;
		case HCI_E_CONNECTION_PACKET_TYPE_CHANGED:
			event_packet_type_changed(hci, skb);
			break;
		case HCI_E_LINK_KEY_NOTIFICATION:
			event_link_key_notification(hci, skb);
			break;
		case HCI_E_INQUIRY_RESULT:
			event_inquiry_result(hci, skb);
			break;
		case HCI_E_AUTHENTICATION_COMPLETE:
			event_authentication_complete(hci, skb);
			break;
		default:
	}
	/* Wakeup Packet Scheduler if transmit is possible ... */
	if (flag)
		netif_schedule(hci->dev);
	hci_deliver_event(hci, skb);		/* for listeners */
	kfree_skb(skb);
	DBFEXIT;
	return err;
}

/*
    Receive acl packet..
*/
int hci_receive_acl(hci_struct *hci, struct sk_buff *skb)
{
	struct sk_buff		*acl;
#ifdef CONFIG_AFFIX_L2CAP
	hci_con			*con;
	HCI_ACL_Packet_Header	*hdr = (void*)skb->data;
#endif

	DBFENTER;
	
	DBPRT(DBCTRL, "HCI ACL packet received\n");
	DBPARSEHCI(DBCTRL, skb->data, skb->len, HCI_ACL, FROM_HOSTCTRL);	
	DBPRTPKT(DBCTRL, skb->data, skb->len);

	if (skb->len < HCI_ACL_HDR_LEN && skb->len < (__btoh16(hdr->Length) + HCI_ACL_HDR_LEN)) {
		BTDEBUG("ACL packet too small, drop it...\n");
		kfree_skb(skb);
		return -EINVAL;
	}

#ifdef CONFIG_AFFIX_L2CAP
	con = hcc_lookup_chandle(hci, HCI_HANDLE(hdr->Connection_Handle));
	if (con == NULL) {	/* No resources available	*/
		DBPRT(DBRECV, "Unknown connection\n");
#ifdef CONFIG_AFFIX_DTL1_FIX
		if (hci->fixit) {
			if (!hci->delayed_acl_skb) {
				BTDEBUG("Delay ACL packet.\n");
				skb_unlink(skb);
				hci->delayed_acl_skb = skb;
				return -EEXIST;
			} else {
				BTDEBUG("Delayed ACL packet already buffered.\n");
				if (hci->delayed_acl_skb != skb)
					kfree_skb(hci->delayed_acl_skb);
				hci->delayed_acl_skb = NULL;
			}
		}
#endif	/* CONFIG_AFFIX_DTL1_FIX */
		kfree_skb(skb);
		return -EINVAL;
	}
#ifdef CONFIG_AFFIX_DTL1_FIX
	if (hci->fixit) {
		if (hci->delayed_acl_skb && hci->delayed_acl_skb != skb) {
			BTDEBUG("Stale delayed ACL packet.\n");
			kfree_skb(hci->delayed_acl_skb);
		}
		hci->delayed_acl_skb = NULL;
	}
#endif /* CONFIG_AFFIX_DTL1_FIX */

	acl = skb_clone(skb, GFP_ATOMIC);
		
	skb_pull(skb, HCI_ACL_HDR_LEN);
	lp_receive_data(con, HCI_PB_FLAG(hdr->Connection_Handle) == HCI_PB_FIRST,
			HCI_BC_FLAG(hdr->Connection_Handle), skb);
	if (acl) {
		hpf_data_ind(con, acl);
	}
	hcc_put(con);
#else  /* CONFIG_AFFIX_L2CAP */
	DBPRT(DBRECV, "Unknown connection\n");
	kfree_skb(skb);
	return -EEXIST;
#endif

	return 0;
}

/*
    Receive sco packet..
*/
int hci_receive_sco(hci_struct *hci, struct sk_buff *skb)
{
	hci_con			*con;
	HCI_SCO_Packet_Header	*hdr = (void*)skb->data;

	DBFENTER;
	
	DBPRT(DBCTRL, "HCI SCO packet received\n");
	DBPARSEHCI(DBCTRL, skb->data, skb->len, HCI_SCO, FROM_HOSTCTRL);	
	DBPRTPKT(DBCTRL, skb->data, skb->len);

	skb_pull(skb, HCI_SCO_HDR_LEN);
	
	if (skb->len < hdr->Length) {
		kfree_skb(skb);
		return -EINVAL;
	}

	con = hcc_lookup_chandle(hci, HCI_HANDLE(hdr->Connection_Handle));
	if (con == NULL) {
		DBPRT(DBRECV, "Unknown connection\n");
		kfree_skb(skb);
		return -EEXIST;
	}
	hpf_data_ind(con, skb);
	hcc_put(con);
	return 0;
}


int hci_receive_data(hci_struct *hci, struct sk_buff *skb)
{
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;

	DBFENTER;

	switch(cb->type) {
	case HCI_ACL:
		hci_receive_acl(hci, skb);
		break;

	case HCI_SCO:
		hci_receive_sco(hci, skb);
		break;

	case HCI_EVENT:
		hci_receive_event(hci, skb);
#ifdef CONFIG_AFFIX_DTL1_FIX
		if (hci->fixit) {
			if (hci->delayed_acl_skb) {
				BTDEBUG("Try to process delayed ACL packet.\n");
				hci_receive_acl(hci, hci->delayed_acl_skb);
			}
		}
#endif
		break;

	default:
		DBPRT(DBRECV, "Unknown HCI type\n");
		goto exit;
	}

	DBFEXIT;
	return 0;
 exit:
	kfree_skb(skb);
	DBFEXIT;
	return -ENODEV;
}



/*
 * **************       HCI COMMAND SUBSYSTEM            *****************
 */
int lp_connect_req(hci_struct *hci, BD_ADDR *bda, hci_con **ret)
{
	hci_con		*con;
	int		err = 0;

	DBFENTER;
	
	con = hcc_create(hci, bda);
	if (!con) {
		return -ENOMEM;
	}
	DBPRT(DBCTRL, "Connection State: %d\n", con->state);
	if (STATE(con) == CON_W4_LDISCREQ)
		ENTERSTATE(con, CON_OPEN);
	else if (STATE(con) == CON_CLOSED) {
		ENTERSTATE(con, CON_W4_LCONREQ);
		con->btdev->paired = 0;
		err = hci_connect_req(con);
		/* start timer here to prevent stale connections */
		hcc_start_timer(con, HCC_CONN_TIMEOUT);
	}

	*ret = con;

	DBFEXIT;
	return err;
}

int lp_add_sco(hci_struct *hci, BD_ADDR *bda, hci_con **ret)
{
	hci_con		*con;
	int		err = 0;

	DBFENTER;
	
	con = hcc_create_sco(hci, bda);
	if (!con) {
		return -ENOMEM;
	}
	if (STATE(con) == CON_CLOSED) {
		ENTERSTATE(con, CON_W4_LCONREQ);
		err = hci_connect_req(con);
		/* start timer here to prevent stale connections */
		hcc_start_timer(con, HCC_CONN_TIMEOUT);
	}

	*ret = con;

	DBFEXIT;
	return err;
}


int lp_disconnect_req(hci_con *con, __u8 reason)
{
	int	err = 0;

	DBFENTER;

	if (STATE(con) == CON_OPEN) {
		ENTERSTATE(con, CON_W4_LDISCREQ);
		err = hci_disconnect_req(con, reason);
		/* start timer here to prevent stale connections */
		hcc_start_timer(con, HCC_DISC_TIMEOUT);
	}

	DBFEXIT;
	return err;
}

int lp_auth_req(hci_con *con)
{
	int	err = 0;

	DBFENTER;

	con->btdev->state = BTDEV_AUTHPENDING;
	hci_auth_req(con);
	hcc_start_timer(con, HCC_AUTH_TIMEOUT);
	
	DBFEXIT;
	return err;
}


/************************ COMMAND LOCK MANAGEMENT ******************************/

hci_lock_t *hci_lock_alloc(void)
{
	hci_lock_t	*lock;

	lock = kmalloc(sizeof(*lock), GFP_ATOMIC);
	if (lock == NULL)
		return NULL;
	memset(lock, 0, sizeof(*lock));

	atomic_set(&lock->refcnt, 0);
	init_MUTEX(&lock->sema);

	return lock;
}

hci_lock_t *hci_lock_lookup(hci_struct *hci, __u16 opcode)
{
	hci_lock_t	*lock;

	btl_read_lock(&hci->hci_locks);	/* lock */

	btl_for_each (lock, hci->hci_locks)
		if (lock->opcode == opcode)
			break;

	if (lock)
		hci_lock_hold(lock);		/* use it */

	btl_read_unlock(&hci->hci_locks);	/* unlock */

	return lock;
}

/*
  create new HCI lock for specified command
*/
hci_lock_t *hci_lock_create(hci_struct *hci, __u16 opcode)
{
	hci_lock_t	*lock;

	btl_write_lock(&hci->hci_locks);	/* lock */
	btl_for_each (lock, hci->hci_locks)
		if (lock->opcode == opcode)
			break;
	if (lock == NULL)  {
		lock = hci_lock_alloc();
		if (lock) {
			lock->hci = hci;
			lock->opcode = opcode;
			__btl_add_tail(&hci->hci_locks, lock);
		}
	}
	hci_lock_hold(lock);		/* use it */
	btl_write_unlock(&hci->hci_locks);	/* unlock */

	return lock;
}

void hci_lock_destroy(hci_lock_t *lock)
{
	hci_struct	*hci = lock->hci;

	btl_write_lock(&hci->hci_locks);	/* lock */
	if (atomic_read(&lock->refcnt) == 0) {
		__btl_unlink(&hci->hci_locks, lock);
		hci_lock_free(lock);
	}
	btl_write_unlock(&hci->hci_locks);	/* unlock */
}


/*
    Bluetooth network device library
*/

void hci_rx_task(unsigned long arg)
{
	hci_struct	*hci = (hci_struct *)arg;
	struct sk_buff	*skb;

	DBFENTER;
	while ((skb = skb_dequeue(&hci->rx_queue))) {
		hci_receive_data(hci, skb);
	}
	DBFEXIT;
}

/* receiving side */
#ifdef CONFIG_AFFIX_OLDRX
int hci_receive(struct sk_buff *skb, struct net_device *dev, struct packet_type *ptype)
{
	hci_struct	*hci = DEV_HCI(dev);

	skb_pull(skb, 1);
	skb->dev = dev;
	hci_receive_data(hci, skb);

	return 0;
}

struct packet_type bluetooth_packet_type = {
	type:	__constant_htons(ETH_P_BLUETOOTH),
	func:	hci_receive
};
#endif

void hci_cleanup(hci_struct *hci)
{
	hci_con		*con, *next;
	int		state;

	DBFENTER;

	btl_write_lock(&hcicons);
	btl_for_each_next (con, hcicons, next) {
		if (con->hci == hci) {
			hcc_hold(con);
			state = SETSTATE(con, DEAD);
			hcc_stop_timer(con);
			switch (state) {
				case CON_OPEN:
#ifdef CONFIG_AFFIX_L2CAP
					lp_disconnect_ind(con);
#endif
					hpf_disconnect_ind(con);
					break;
				default:
#ifdef CONFIG_AFFIX_L2CAP
					lp_connect_cfm(con, HCI_ERR_HARDWARE_FAILURE);
#endif
					hpf_connect_cfm(con, HCI_ERR_HARDWARE_FAILURE);
			}
			__hcc_put(con);
		}
	}
	btl_write_unlock(&hcicons);
	DBFEXIT;
}

static inline void hci_initdev(struct net_device *dev)
{
	DBFENTER;
	hci_create(dev);
	hci_sched_init(dev);	/* set dev->qdisc */
	hcidev_state_change(dev, HCIDEV_REGISTER);
	DBFEXIT;
}

static inline void hci_exitdev(struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);
	hci_listener_t	*evl, *n;

	DBFENTER;
	netif_device_detach(dev);	// do detach	
	hcidev_state_change(dev, HCIDEV_UNREGISTER);
	btl_write_lock(&listeners);
	btl_for_each_next (evl, listeners, n) {
		evl->err = -ENODEV;
		if (evl->handler)
			evl->handler(evl, event, NULL);

		if (waitqueue_active(&evl->sleep))
			wake_up_interruptible(&evl->sleep);
	}
	btl_write_unlock(&listeners);
	hci_put(hci);
	DBFEXIT;
}

static inline void hci_updev(struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);

	DBFENTER;

	hci->hdrlen = dev->hard_header_len?dev->hard_header_len:1;
	hci->cmd_count = 1;		/* inititally */
	tasklet_hi_enable(&hci->rx_task);
	hci->flags = 0;

	DBFEXIT;
}

static inline void hci_downdev(struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);

	DBFENTER;

	tasklet_disable(&hci->rx_task);
	hci->flags = 0;
	hcidev_state_change(dev, HCIDEV_DOWN);
	hci_cleanup(hci);

	DBFEXIT;
}


/* 
   Called when a NEW network device is available
*/
int bluetooth_event(struct notifier_block *nb, unsigned long event, void *arg)
{
	struct net_device	*dev = arg;

	DBFENTER;

	if (dev->type != ARPHRD_BLUETOOTH)
		return NOTIFY_DONE; /* I am not interested in other dev */

	switch (event) {
		case NETDEV_REGISTER:
			DBPRT(DBINIT, "Bluetooth device registered\n");
			hci_initdev(dev);
			break;
		case NETDEV_UNREGISTER:
			DBPRT(DBINIT, "Bluetooth device unregistered\n");
			hci_exitdev(dev);
			break;
		case NETDEV_UP:
			DBPRT(DBINIT, "Bluetooth device up\n");
			hci_updev(dev);
			break;
		case NETDEV_DOWN:
			DBPRT(DBINIT, "Bluetooth device down\n");
			hci_downdev(dev);
			break;
		default:
			break;
	}
	DBFEXIT;
	return NOTIFY_DONE;
}

void hcidev_state_change(struct net_device *dev, int event)
{
	hci_struct	*hci = DEV_HCI(dev);

	/* pass to kernel */
	notifier_call_chain(&affix_chain, event, hci);
	/* local chain handler */
	switch (event) {
		case HCIDEV_DETACH:
			rtnl_lock();
			dev_close(dev);
			rtnl_unlock();
			hci_run_hotplug(hci, "detach");
			break;
		case HCIDEV_ATTACH:
			hci_run_hotplug(hci, "attach");
			break;
		default:
	}
	/* pass to processes */
	hci_state_change(hci, event);
}

struct notifier_block bluetooth_notifier_block = {
	notifier_call:	bluetooth_event,
};


int affix_register_notifier(struct notifier_block *nb)
{
	return notifier_chain_register(&affix_chain, nb);
}

int affix_unregister_notifier(struct notifier_block *nb)
{
	return notifier_chain_unregister(&affix_chain, nb);
}

/* 
   called from drivers like btuart, btusb
*/
struct net_device *hcidev_alloc(void)
{
	struct net_device	*dev;
	int 			alloc_size;

	/* ensure 32-byte alignment of the private area */
	alloc_size = sizeof(*dev) + sizeof(hci_struct) + 31;
	dev = (struct net_device *) kmalloc(alloc_size, in_softirq() ? GFP_ATOMIC : GFP_KERNEL);
	if (dev == NULL) {
		BTERROR("Unable to allocate device memory.\n");
		return NULL;
	}
	memset(dev, 0, alloc_size);

	if (dev_alloc_name(dev, "bt%d") < 0) {
		kfree(dev);
		return NULL;
	}
	dev->features |= NETIF_F_DYNALLOC;
	dev->type = ARPHRD_BLUETOOTH;
	dev->mtu = 1500;
	dev->addr_len = ETH_ALEN;
	dev->tx_queue_len = 10;
	dev->flags = 0;	//IFF_BROADCAST|IFF_MULTICAST;
	memset(dev->broadcast,0xFF, ETH_ALEN);
	dev_init_buffers(dev);
	return dev;
}

int hcidev_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);
	hci_skb_cb	*cb;
	int		err;

	if (!skb)
		return 0;
	cb = (hci_skb_cb *)skb->cb;
	skb->pkt_type = cb->type;
	skb_pull(skb, 1);	/* remove uart byte */
	err = hci->hard_start_xmit(skb, dev);
	if (err) { // skb not alive
		*skb_push(skb, 1) = skb->pkt_type;	/* UART typing */
		cb->type = skb->pkt_type;
	}
	return err;
}

int hcidev_register(struct net_device *dev, void *param)
{
	hci_struct	*hci = DEV_HCI(dev);
	int		err;
	
	rtnl_lock();
	hci->hard_start_xmit = dev->hard_start_xmit;
	dev->hard_start_xmit = hcidev_hard_start_xmit;
	err = register_netdevice(dev);
	rtnl_unlock();
	hci->devinfo = param;
	hci_run_hotplug(hci, "register");

	return err;
}

void hcidev_unregister(struct net_device *dev)
{
	hci_struct	*hci = DEV_HCI(dev);
	
	hci_run_hotplug(hci, "unregister");
	hci->devinfo = NULL;
	rtnl_lock();
	unregister_netdevice(dev);
	rtnl_unlock();
}


/*
 * UART related
 */
int affix_set_uart(struct affix_uart_operations *ops)
{
	DBFENTER;
	
	if (ops == NULL)
		affix_uart_ops.owner = NULL;
	else
		affix_uart_ops = *ops;
	
	DBFEXIT;
	return 0;
}

int affix_open_uart(struct open_uart *line)
{
	affix_uart_t	uart;
	int		err;

	DBFENTER;

	if (!affix_uart_ops.owner || !try_inc_mod_count(affix_uart_ops.owner)) { 
		request_module("affix_uart");
		if (!affix_uart_ops.owner || !try_inc_mod_count(affix_uart_ops.owner))
			return -ENODEV;
	}
	memset(&uart, 0, sizeof(uart));
	strncpy(uart.name, line->dev, AFFIX_UART_PATHLEN);
	uart.prodid = line->type;
	uart.proto = line->proto;
	uart.speed = line->speed;
	uart.flags = line->flags;

	err = affix_uart_ops.attach(&uart);
	if (!err)
		return 0;	// return if no error
	if (affix_uart_ops.owner)
		__MOD_DEC_USE_COUNT(affix_uart_ops.owner);
	DBFEXIT;
	return err;
}

int affix_close_uart(struct open_uart *line)
{
	int	err;

	DBFENTER;
	
	if (!affix_uart_ops.owner)
		return -ENODEV;
		
	err = affix_uart_ops.detach(line->dev);
	if (err)
		return err;
	
	if (affix_uart_ops.owner)
		__MOD_DEC_USE_COUNT(affix_uart_ops.owner);

	DBFEXIT;
	return 0;
}

char *hci_devtype(int type)
{
	switch (type) {
		case HCI_USB:
			return "usb";
		case HCI_PCCARD:
			return "pcmcia";
		case HCI_UART:
			return "uart";
		case HCI_UART_CS:
			return "uart_cs";
		default:
			return "unknown";
	}
}

static int uart_speed(int s)
{
	switch (s) {
	case B9600:
		return 9600;
	case B19200:
		return 19200;
	case B38400:
		return 38400;
	case B57600:
		return 57600;
	case B115200:
		return 115200;
	case B230400:
		return 230400;
	case B460800:
		return 460800;
	case B921600:
		return 921600;
	case B1000000:
		return 1000000;
	default:
		return 57600;
	}
}

void hci_run_hotplug(hci_struct *hci, char *verb)
{
	char 	*argv [6], **envp, *buf, *scratch;
	int	i = 0, value;

	DBFENTER;
	
	if (in_interrupt ()) {
		DBPRT(DBINIT, "In_interrupt");
		return;
	}

	if (!current->fs->root) {
		/* statically linked USB is initted rather early */
		DBPRT(DBINIT, "%s -- no FS yet", verb);
		return;
	}
	
	if (!(envp = (char **) kmalloc (20 * sizeof (char *), in_softirq() ? GFP_ATOMIC : GFP_KERNEL))) {
		DBPRT(DBINIT, "enomem");
		return;
	}
	
	if (!(buf = kmalloc (256, in_softirq() ? GFP_ATOMIC : GFP_KERNEL))) {
		kfree (envp);
		DBPRT(DBINIT, "enomem2");
		return;
	}

	/* only one standardized param to hotplug command: type */
	argv [0] = "/etc/affix/hwinit";
	argv [1] = "affix";	/* hotplug compliant */
	argv [2] = hci->name;
	argv [3] = verb;
	argv [4] = hci_devtype(hci->dev->if_port);
	argv [5] = 0;

	/* minimal command environment */
	envp [i++] = "HOME=/";
	envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";

#ifdef	DEBUG
	/* hint that policy agent should enter no-stdout debug mode */
	envp [i++] = "DEBUG=kernel";
#endif
	/* extensible set of named bus-specific parameters,
	 * supporting multiple driver selection algorithms.
	 */
	scratch = buf;

#if 0
	/* action:  add, remove */
	envp [i++] = scratch;
	scratch += sprintf (scratch, "ACTION=%s", verb) + 1;

	envp [i++] = scratch;
	scratch += sprintf (scratch, "IFACE=%s", hci->name) + 1;
#endif

	if (!hci->devinfo)
		goto skip;

#if defined(CONFIG_USB)  || defined(CONFIG_USB_MODULE)
	if (hci->dev->if_port == HCI_USB) {
		struct usb_device	*usb_dev = hci->devinfo;

		if (usb_dev->devnum < 0) {
			DBPRT(DBINIT, "device already deleted ??");
			kfree (envp);
			kfree (argv);
			return;
		}

#ifdef	CONFIG_USB_DEVICEFS
		/* 
		 * If this is available, userspace programs can directly read
		 * all the device descriptors we don't tell them about.  Or
		 * even act as usermode drivers.
		 *
		 */
		envp [i++] = "DEVFS=/proc/bus/usb";
		envp [i++] = scratch;
		scratch += sprintf (scratch, "DEVICE=/proc/bus/usb/%03d/%03d",
				usb_dev->bus->busnum, usb_dev->devnum) + 1;
#endif
		/* per-device configuration hacks are common */
		envp [i++] = scratch;
		scratch += sprintf (scratch, "PRODUCT=%x/%x/%x",
				usb_dev->descriptor.idVendor,
				usb_dev->descriptor.idProduct,
				usb_dev->descriptor.bcdDevice) + 1;

		/* class-based driver binding models */
		envp [i++] = scratch;
		scratch += sprintf (scratch, "TYPE=%d/%d/%d",
				usb_dev->descriptor.bDeviceClass,
				usb_dev->descriptor.bDeviceSubClass,
				usb_dev->descriptor.bDeviceProtocol) + 1;
	} else 
#endif
	if (hci->dev->if_port == HCI_UART || hci->dev->if_port == HCI_UART_CS) {
		affix_uart_t	*uart = hci->devinfo;

		envp [i++] = scratch;
		scratch += sprintf (scratch, "DEVICE=%s", uart->name) + 1;
		envp [i++] = scratch;
		scratch += sprintf (scratch, "PRODUCT=%#x:%#x", uart->manfid, uart->prodid) + 1;
		envp [i++] = scratch;
		scratch += sprintf (scratch, "SPEED=%d", uart_speed(uart->speed)) + 1;
		envp [i++] = scratch;
		scratch += sprintf (scratch, "FLAGS=%#x", uart->flags) + 1;
	}

skip:
	envp [i++] = 0;
	/* assert: (scratch - buf) < sizeof buf */

	/* NOTE: user mode daemons can call the agents too */

	//DBPRT(DBINIT, "%s %s %d", argv [0], verb, dev->devnum);
	value = call_usermodehelper (argv[0], argv, envp);
	kfree (buf);
	kfree (envp);
	if (value != 0)
		DBPRT(DBINIT, "returned 0x%x", value);

	DBFEXIT;
}

/* 
   Socket interface subsystem
*/

struct net_proto_family		*btprotos[BTPROTO_MAX];

int affix_sock_register(struct net_proto_family *pf, int protocol)
{
	if (protocol >= BTPROTO_MAX)
		return -EINVAL;

	if (btprotos[protocol])
		return -EEXIST;
	
	btprotos[protocol] = pf;
	
	return 0;
}

int affix_sock_unregister(int protocol)
{
	if (protocol >= BTPROTO_MAX)
		return -EINVAL;

	btprotos[protocol] = NULL;

	return 0;
}

int affix_sock_create(struct socket *sock, int protocol)
{
	DBFENTER;
	if (protocol >= BTPROTO_MAX)
		return -EPROTONOSUPPORT;

	if (btprotos[protocol] == NULL) {
		if (protocol == BTPROTO_RFCOMM)
			request_module("affix_rfcomm");
	}
	if (btprotos[protocol])
		return btprotos[protocol]->create(sock, protocol);

	return -EPROTONOSUPPORT;
}


struct net_proto_family bluetooth_family_ops = {
	family:	PF_AFFIX,
	create:	affix_sock_create
};


extern long acl_skb_count;
/*
    PROC file system
*/
int hci_device_proc_read(char *buf, char **start, off_t offset, int len)
{
	int 		count = 0, i = 0;
	hci_struct	*hci;
	hci_con		*con;

	DBFENTER;

	count += sprintf(buf+count, "Overall ACL queue length: %ld\n", acl_skb_count);

	count += sprintf(buf+count, "Devices (%d)\n", hcidevs.len);
	btl_read_lock(&hcidevs);
	btl_for_each (hci, hcidevs) {
		count += sprintf(buf+count, "%s (%d): %s, flags: %#x, cmd_count: %d, acl_count: %d\n", 
				hci->name, hci->hd, bda2str(&hci->bda), hci->flags, hci->cmd_count, hci->acl_count);
	}
	btl_read_unlock(&hcidevs);

	count += sprintf(buf+count, "Connections (%d)\n", hcicons.len);
	btl_read_lock(&hcicons);
	btl_for_each (con, hcicons) {
		i++;
		count += sprintf(buf+count, "%d: handle: %#x, bda: %s, type: %s, pkt_type: %#x\n   qlen: %d, refcnt: %d, state: %d\n",
				i, con->chandle, bda2str(&con->bda), (con->link_type == HCI_LT_ACL)?"ACL":"SCO", con->pkt_type, skb_queue_len(&con->tx_queue), atomic_read(&con->refcnt), STATE(con));
	}
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return count;
}

struct hci_proc_entry {
	char		*name;
	get_info_t	*fn;
};

struct hci_proc_entry dir[] = {
	{"hci", hci_device_proc_read},
	{0, 0}
};

int hci_proc_init(void)
{
	struct proc_dir_entry	*ent;
	struct hci_proc_entry	*p;

	ent = create_proc_entry("affix", S_IFDIR, proc_net);
	if (!ent)
		return -1;

	proc_affix = ent;

	for (p = dir; p->name; p++)
		ent = create_proc_info_entry(p->name, 0, proc_affix, p->fn);
	
	return 0;
}

void hci_proc_exit(void)
{
	struct hci_proc_entry	*p;
	
	if (proc_affix == NULL)
		return;

	for (p = dir; p->name; p++)
		remove_proc_entry(p->name, proc_affix);

	proc_affix = NULL;
	remove_proc_entry("affix", proc_net);
}


int init_hpf(void);
int exit_hpf(void);

/* 
   main hci initialization 
*/
int __init init_hci(void)
{
	int		err = -ENOMEM;

	DBFENTER;

	btl_head_init(&hcidevs);
	btl_head_init(&hcicons);
	btl_head_init(&listeners);
	btl_head_init(&hcimgrs);

	affix_set_uart(NULL);

	/* register our socket family */
	err = sock_register(&bluetooth_family_ops);
	if (err)
		goto err1;

	err = hci_proc_init();
	if (err < 0) {
		BTERROR("Unable to initialize procfs\n");
		goto err2;
	}

	err = init_hpf();
	if (err < 0) {
		BTERROR("Unable to intialize HCI socket interface\n");
		goto err3;
	}
#ifdef CONFIG_AFFIX_OLDRX
	dev_add_pack(&bluetooth_packet_type);
#endif
	register_netdevice_notifier(&bluetooth_notifier_block);

	/* run it at last when all stuff initialized */
	err = hci_start_manager();
	if (err < 0) {
		BTERROR("Unable to start hci thread\n");
		goto err4;
	}

	DBFEXIT;
	return 0;
err4:
	exit_hpf();
err3:
	hci_proc_exit();	
err2:
	sock_unregister(PF_AFFIX);
err1:
	return err;
}

void exit_hci(void)
{
	DBFENTER;

	hci_stop_manager();
	exit_hpf();
	hci_proc_exit();
	unregister_netdevice_notifier(&bluetooth_notifier_block);
#ifdef CONFIG_AFFIX_OLDRX
	dev_remove_pack(&bluetooth_packet_type);
#endif
	sock_unregister(PF_AFFIX);

	DBFEXIT;
}


int init_debug(void);

int init_l2cap(void);
void exit_l2cap(void);

int init_rfcomm(void);
void exit_rfcomm(void);


/*
  affix_init
  
  This function register new line discipline for BT
  
  returns: zero on success, non-zero on failure.
*/
int __init affix_init(void)
{
	int	err;
	DBFENTER;

#if defined(CONFIG_AFFIX_DEBUG)
	err = init_debug();
	if (err)
		goto err1;
#endif
	
	err = init_hci();
	if (err)
		goto err1;

#if defined(CONFIG_AFFIX_L2CAP)
	err = init_l2cap();
	if (err)
		goto err2;

#if defined(CONFIG_AFFIX)	/* inside kernel */
#if defined(CONFIG_AFFIX_RFCOMM)
	err = init_rfcomm();
	if (err)
		goto err3;
#endif	/*RFCOMM*/
#if defined(CONFIG_AFFIX_PAN)
	err = init_pan();
	if (err)
		goto err4;
#endif	/*PAN*/
#endif	/* kernel */

#endif  /*L2CAP*/

	printk("Affix Bluetooth Protocol Stack loaded\n");
	printk("Copyright (C) 2001, 2002 Nokia Corporation\n");
	printk("Written by Dmitry Kasatkin <dmitry.kasatkin@nokia.com>\n");

	DBFEXIT;
	return err;

#if defined(CONFIG_AFFIX_L2CAP)
#if defined(CONFIG_AFFIX)	/* kernel */
#if defined(CONFIG_AFFIX_PAN)
 err4:
#if defined(CONFIG_AFFIX_RFCOMM)
	exit_rfcomm();
#endif
#endif
#if defined(CONFIG_AFFIX_RFCOMM)
 err3:
#endif
#if defined(CONFIG_AFFIX_RFCOMM) || defined(CONFIG_AFFIX_PAN)
	exit_l2cap();
#endif
#endif	/* kernel */
 err2:
	exit_hci();
#endif /*L2CAP*/
 err1:
	return err;
}

/*
  affix_exit

  returns: nothing
*/
void __exit affix_exit(void)
{
	DBFENTER;
	
#if defined(CONFIG_AFFIX_L2CAP)
#if defined(CONFIG_AFFIX)
#if defined(CONFIG_AFFIX_PAN)
	exit_pan();
#endif
#if defined(CONFIG_AFFIX_RFCOMM)
	exit_rfcomm();
#endif
#endif
	exit_l2cap();
#endif
	exit_hci();

	DBFEXIT;
}

EXPORT_SYMBOL(affix_dbmask);
EXPORT_SYMBOL(proc_affix);
EXPORT_SYMBOL(hcidev_alloc);
EXPORT_SYMBOL(hcidev_register);
EXPORT_SYMBOL(hcidev_unregister);
EXPORT_SYMBOL(hcidev_state_change);
EXPORT_SYMBOL(affix_set_uart);
EXPORT_SYMBOL(affix_set_pan);
EXPORT_SYMBOL(affix_register_notifier);
EXPORT_SYMBOL(affix_unregister_notifier);
EXPORT_SYMBOL(affix_sock_register);
EXPORT_SYMBOL(affix_sock_unregister);
EXPORT_SYMBOL(hci_lookup_bda);
EXPORT_SYMBOL(hci_select);
EXPORT_SYMBOL(lp_auth_req);

EXPORT_SYMBOL(hci_run_hotplug);


#ifdef MODULE
/*  If we are resident in kernel we call affix_init manually after root FS is mounted.  */
module_init(affix_init);
module_exit(affix_exit);

MODULE_AUTHOR("Dmitry Kasatkin <dmitry.kasatkin@nokia.com>");
MODULE_DESCRIPTION("BTCORE (HCI, L2CAP, ...) Bluetooth Driver Core for Linux");
MODULE_PARM(affix_dbmask, "i"); /* For insmod */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

#endif // MODULE


/*
  NOTES:

  4. maybe HUP signal restart mgr_thread()
  6. Remove listener in all functions in hci_cmd.c in case of error

  - btdev_execute_command & btdev_recv_event should check hci->state
  - hci->hwtimeout. HCI manager set it to try command execution  - 20sec now
  - check btdev->evl->hci. It's not used for some reason
*/


/*
 * tasklet_disable
 * tasklet_enable
 * tasklet_kill
 * tasklet_schedule
 * tasklet_hi_schedule
 *
 */
