/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: hci_sched.c,v 1.10 2002/11/13 00:26:01 kds Exp $

   HCI Packet Scheduler

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <asm/bitops.h>

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <net/pkt_sched.h>

/* Local Includes */
#define FILEBIT	DBHCISCHED

#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>


long acl_skb_count;


/* Queue discipline for HCI */

int hci_sched_enqueue(struct sk_buff *skb, struct Qdisc * qdisc)
{
	hci_skb_cb		*cb = (hci_skb_cb*)skb->cb;
	struct net_device	*dev = qdisc->dev;
	hci_struct		*hci = DEV_HCI(dev);
	hci_con			*con = cb->con;

	DBFENTER;

	switch (cb->type) {
	case HCI_ACL:
		skb_queue_tail(&con->tx_queue, skb);
		acl_skb_count++;
		break;
	case HCI_SCO:
		skb_queue_tail(&con->tx_queue, skb);
		break;
	case HCI_COMMAND:
		skb_queue_tail(&hci->cmd_queue, skb);
		break;
	default:
		DBPRT(DBXMITE, "Unknown packet type: %d\n", cb->type);
	}
	
	DBFEXIT;
	return 0;
}

/*
   this function is called if the ncc_xmit returns none-zero value..
   I should queue packet back 
   and check dev status to do not send again from hci_dequeue
*/
int hci_sched_requeue(struct sk_buff *skb, struct Qdisc* qdisc)
{
	hci_skb_cb		*cb = (hci_skb_cb*)skb->cb;
	struct net_device	*dev = qdisc->dev;
	hci_struct		*hci = DEV_HCI(dev);
	hci_con			*con;
	HCI_ACL_Packet_Header	*hdr;


	DBFENTER;

	skb_pull(skb, 1);	/* UART typing */

	switch (cb->type) {
	case HCI_ACL:
		hdr = (void*)skb->data;
		con = hcc_lookup_chandle(hci, HCI_HANDLE(hdr->Connection_Handle));
		if (con == NULL) {
			kfree_skb(skb);
			return -ENOTCONN;
		}
		/* Remove ACL header */
		skb_pull(skb, HCI_ACL_HDR_LEN);
		skb_queue_head(&con->tx_queue, skb);
		con->prior--;
		hci->acl_count++;
		acl_skb_count++;
		hcc_put(con);
		break;
	case HCI_SCO:
		hdr = (void*)skb->data;
		con = hcc_lookup_chandle(hci, HCI_HANDLE(hdr->Connection_Handle));
		if (con == NULL) {
			kfree_skb(skb);
			return -ENOTCONN;
		}
		skb_queue_head(&con->tx_queue, skb);
		con->prior--;
		hci->sco_count++;
		hcc_put(con);
		break;
	case HCI_COMMAND:
		skb_queue_head(&hci->cmd_queue, skb);
		hci->cmd_count++;
		break;
	default:
		DBPRT(DBXMITE, "Unknown packet type: %d\n", cb->type);
	}

	if (net_ratelimit())
		printk(KERN_DEBUG "%s deferred output. It is buggy.\n", skb->dev->name);


	DBFEXIT;
	return 0;
}

/*
   TRUE, if t1<t2
*/
int inline cmp_timeval(struct timeval *t1, struct timeval *t2)
{
	if (t1->tv_sec < t2->tv_sec) return 1;
	else if (t1->tv_sec > t2->tv_sec) return 0;
	else if (t1->tv_usec < t2->tv_usec) return 1;
	else return 0;
}

/*
   select candidate connection to be sended from
*/
struct sk_buff *select_packet(hci_struct *hci, __u8 type, hci_con **ret)
{
	hci_con		*con;
	struct sk_buff	*rskb = NULL, *skb;

	DBFENTER;

	*ret = NULL;

	btl_read_lock(&hcicons);	//lock

	/* find first */
	btl_for_each (con, hcicons)
		if( con->hci == hci && con->link_type == type &&
		    (skb = skb_peek(&con->tx_queue)) != NULL ) {
			*ret = con;
			rskb = skb;
			break;
		}
	
	if (rskb == NULL) /* nothing */
		goto exit;

	/* find best */
	btl_for_each_cur(con, hcicons)
		if( con->hci == hci && con->link_type == type &&
		    (skb = skb_peek(&con->tx_queue)) != NULL ) {
			if (con->prior == (*ret)->prior) {
				if (cmp_timeval(&skb->stamp, &rskb->stamp)) {
					*ret = con;
					rskb = skb;
				}
			} else if (con->prior < (*ret)->prior) {
				*ret = con;
				rskb = skb;
			}
		}

	if (*ret)
		hcc_hold(*ret);
exit:
	btl_read_unlock(&hcicons);	//unlock
	DBFEXIT;
	return rskb;
}

struct sk_buff *dequeue_acl_packet(hci_con *con, struct sk_buff *skb)
{
	struct sk_buff		*s;
	hci_skb_cb		*skb_cb, *s_cb;
	int			len, pb;
	HCI_ACL_Packet_Header	*hdr;
	hci_struct		*hci = con->hci;

	skb_cb = (hci_skb_cb*)skb->cb;
	pb = skb_cb->pb;

	if (skb->len <= con->mtu) {
		skb_unlink(skb);
		s = skb;
		len = s->len;
		acl_skb_count--;
	} else {
		len = con->mtu;
		s = alloc_skb(con->hdrlen + len, GFP_ATOMIC);
		if (s == NULL) {
			BTERROR("No memory\n");
			return NULL;
		}
		skb_reserve(s, con->hdrlen);
		memcpy(skb_put(s, len), skb->data, len);

		s_cb = (hci_skb_cb*)s->cb;
		s_cb->type = HCI_ACL;
		s_cb->pb = pb;
		s->stamp = skb->stamp;
		s->dev = hci->dev;

		skb_pull(skb, len);
		skb_cb->pb = HCI_PB_MORE;	/* update current Packet Boundary */
	}

	hdr = (void*)skb_push(s, HCI_ACL_HDR_LEN);
	hdr->Connection_Handle = __htob16(con->chandle | pb | HCI_BC_PP);
	hdr->Length = __htob16(len);

	return s;
}

struct sk_buff *dequeue_sco_packet(hci_con *con, struct sk_buff *skb)
{
	struct sk_buff		*s;
	hci_skb_cb		*skb_cb, *s_cb;
	int			len;
	HCI_SCO_Packet_Header	*hdr;
	hci_struct		*hci = con->hci;

	skb_cb = (hci_skb_cb*)skb->cb;

	if (skb->len <= con->mtu) {
		skb_unlink(skb);
		s = skb;
		len = s->len;
	} else {
		len = con->mtu;
		s = alloc_skb(con->hdrlen + len, GFP_ATOMIC);
		if (s == NULL) {
			BTERROR("No memory\n");
			return NULL;
		}
		skb_reserve(s, con->hdrlen);
		memcpy(skb_put(s, len), skb->data, len);

		s_cb = (hci_skb_cb*)s->cb;
		s_cb->type = HCI_SCO;
		s->stamp = skb->stamp;
		s->dev = hci->dev;
		skb_pull(skb, len);
	}

	hdr = (void*)skb_push(s, HCI_SCO_HDR_LEN);
	hdr->Connection_Handle = __htob16(con->chandle);
	hdr->Length = len;

	return s;
}

struct sk_buff *hci_sched_dequeue(struct Qdisc * qdisc)
{
	struct net_device	*dev = qdisc->dev;
	hci_struct		*hci = DEV_HCI(dev);
	struct sk_buff		*acl, *sco, *cmd, *r;
	hci_skb_cb		*cb;
	hci_con			*acl_con = NULL, *sco_con = NULL;

	DBFENTER;

	DBPRT(DBXMIT, "cmd_count: %d, acl_count: %d\n", hci->cmd_count, hci->acl_count);

	/* now select the element from queues */
	cmd = (hci->cmd_count) ? skb_peek(&hci->cmd_queue) : NULL;
	acl = (hci->acl_count) ? select_packet(hci, HCI_LT_ACL, &acl_con) : NULL;
	sco = (hci->sco_count) ? select_packet(hci, HCI_LT_SCO, &sco_con) : NULL;

	if (cmd) {
		if (acl)
			r = (cmp_timeval(&cmd->stamp, &acl->stamp)?cmd:acl);
		else
			r = cmd;
	} else {
		if (acl)
			r = acl;
		else {
			r = sco;	// cmd == NULL and acl == NULL
			if (r == NULL)
				return NULL;
			goto exit;
		}
	}

	if (sco)
		r = (cmp_timeval(&r->stamp, &sco->stamp)?r:sco);

 exit:
	cb = (hci_skb_cb*)r->cb;

	switch (cb->type) {
	case HCI_COMMAND:
		skb_unlink(r);
		hci->cmd_count--;
		break;

	case HCI_ACL:
		r = dequeue_acl_packet(acl_con, r);
		if (r == NULL)
			break;
		acl_con->prior++;
		hci->acl_count--;
		break;

	case HCI_SCO:
	/*
		r = dequeue_sco_packet(sco_con, r);
		if (r == NULL)
			break;
	*/
		skb_unlink(r);
		sco_con->prior++;
		hci->sco_count--;
		break;
	}

	if (acl_con)
		hcc_put(acl_con);
	if (sco_con)
		hcc_put(sco_con);

	*skb_push(r, 1) = cb->type;	/* UART typing */
	r->protocol = __constant_htons(ETH_P_BLUETOOTH);
		
	DBFEXIT;
	return r;
}

struct Qdisc_ops hci_sched_qdisc_ops =
{
	NULL,
	NULL,
	"affix",
	0,
	hci_sched_enqueue,
	hci_sched_dequeue,
	hci_sched_requeue,
};

void hci_sched_init(struct net_device *dev)
{
	DBFENTER;
	
	if (dev->qdisc_sleeping == &noop_qdisc) {
		if (dev->tx_queue_len) {
			struct Qdisc	*qdisc;

			DBPRT(DBINIT, "Activating HCI packet scheduler\n");
			qdisc = qdisc_create_dflt(dev, &hci_sched_qdisc_ops);
			if (qdisc == NULL) {
				BTERROR("Activation failed: no mem\n");
				return;
			}
			write_lock(&qdisc_tree_lock);
			dev->qdisc_sleeping = qdisc;
			write_unlock(&qdisc_tree_lock);
			
			dev->watchdog_timeo = 5*HZ;
			// dev->tx_timeout(dev) what is it???
		}
	}
	
	DBFEXIT;
}


