/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btuart.h,v 1.24 2002/11/12 10:41:01 kds Exp $

   BTUART - physical protocol layer for bluetooth uart devices

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/

#ifndef _BTUART_H
#define _BTUART_H

#define BTUART_DEVNAME	"affix_uart"

#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include <affix/hci.h>

#define N_BTUART	14	/* line discipline */

/* NOKIA */
#define MANFID_NOKIA		0x0124
#define PRODID_DTL1		0x0900
#define PRODID_DTL4		0x1000

/* SOCKET */
#define MANFID_SOCKET		0x0104
#define PRODID_SOCKET_BTCF1	0x009f



typedef struct btuart_t	btuart_t;

typedef struct {
	int	proto;
	int	hard_hdr_len;
	int	(*enqueue)(btuart_t *btuart, struct sk_buff *skb);
	struct sk_buff*	(*dequeue)(btuart_t *btuart);
	int	(*init)(btuart_t *btuart);
	void	(*uninit)(btuart_t *btuart);
	void	(*recv_buf)(btuart_t *, const unsigned char *, int);
} btuart_proto_t;

/* control structure for transport driver */
struct btuart_t {
	struct list_head	q;	/* for queueing */

	/* network device subsystem */
	struct net_device_stats	stats;
	struct net_device	*dev;

#define BTUART_RUNNING		0
#define BTUART_XMIT_WAKEUP	1
	int			flags;
	spinlock_t		xmit_lock;	/* xmit lock */
	rwlock_t		lock;
	
	/* physical/tty stuff */
	affix_uart_t		uart;
	kdev_t			kdev;
	btuart_proto_t		proto;
	struct file		*filp;		/* for tty driver */
	struct tty_struct 	*tty;

	/* tx state */
	struct sk_buff_head	tx_q;
	struct sk_buff		*tx_skb;	/* sended packet */
	struct tasklet_struct	tx_task;

	/* rx state */
	int			pkt_len;
	struct sk_buff		*rx_skb;
	int			hdr_len;
	int			rx_count;
	__u8			hdr[5];
	
#if defined(CONFIG_AFFIX_UART_TLP)
	int			flowmask;	/* tlp flow mask */
	int			sync;
	struct timeval		sync_stamp;
	int			uart_ri_latch;
#ifdef CONFIG_AFFIX_DEBUG
	int			irqcnt;
#endif
#endif
	/* protocol private */
	void			*priv;
};

/* H4 */
#if defined(CONFIG_AFFIX_UART_H4)
int  btuart_init_h4(btuart_t *btuart);
void  btuart_uninit_h4(btuart_t *btuart);
void btuart_recv_buf_h4(btuart_t *, const unsigned char *, int);
int btuart_enqueue_h4(btuart_t *btuart, struct sk_buff *skb);
struct sk_buff *btuart_dequeue_h4(btuart_t *btuart);
#endif

#if defined(CONFIG_AFFIX_UART_TLP)
/* support for the Nokia Connectivity Card */

int  btuart_init_tlp(btuart_t *btuart);
void  btuart_uninit_tlp(btuart_t *btuart);
void btuart_recv_buf_tlp(btuart_t *, const unsigned char *, int);
int btuart_enqueue_tlp(btuart_t *btuart, struct sk_buff *skb);
struct sk_buff *btuart_dequeue_tlp(btuart_t *btuart);
/* the structure of the TLP protocol header */
typedef struct {
	__u8	type;
	__u8	zero;
	__u16	length;
} __PACK__ tlp_hdr_t;

#define TLP_HDR_LEN		4

#define TLP_CONTROL		0x80
#define TLP_HCI_COMMAND		0x81
#define TLP_HCI_ACL		0x82
#define TLP_HCI_SCO		0x83
#define TLP_HCI_EVENT		0x84

#define TLP_PAD_LEN(len)	(len & 0x0001)
#define TLP_FCS_LEN		2

#define NCC_SYNC		0x80
#define NCC_SYNC_TIME		(5000)

#define	TLP_CMD_ON		0x0001
#define	TLP_ACL_ON		0x0002
#define	TLP_SCO_ON		0x0004

#endif	/* CONFIG_AFFIX_UART_TLP */

void btuart_receive_buf(void *data);
int btuartld_init(btuart_t *btuart);
void btuartld_uninit(btuart_t *btuart);

int affix_uart_attach(affix_uart_t *serial);
int affix_uart_detach(char *name);
void affix_uart_suspend(char *name);
void affix_uart_resume(char *name);

int kdev_ioctl(struct file *filp, int cmd, void *arg);
void btuart_xmit_wakeup(btuart_t *btuart);

#endif
