/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: bluez.h,v 1.14 2002/11/12 10:51:04 kds Exp $

   BlueZ -> Affix adoptation layer

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/


#ifndef _AFFIX_BLUEZ_H
#define _AFFIX_BLUEZ_H

#ifndef FILEBIT
#define FILEBIT	DBDRV
#endif

#include <linux/netdevice.h>

#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/bluetooth.h>

#define BT_ERR(fmt, args...)   printk(KERN_ERR __FUNCTION__ ": " fmt "\n", ##args)

#ifdef CONFIG_AFFIX_DEBUG
	#define BT_DBG(fmt, args...) \
	{ \
		if ((affix_dbmask & FILEBIT) && (affix_dbmask & (DBCTRL))) { \
			if (affix_dbmask & DBFNAME) \
				printk(KERN_DEBUG __FUNCTION__ ": " fmt "\n", ##args); \
			else \
				printk(KERN_DEBUG fmt "\n", ##args); \
		} \
	}

#else
#define BT_DBG(args...)
#endif /* CONFIG_AFFIX_DEBUG */

/* HCI device states */
enum hci_dev_states_t
{
	HCI_INIT = 0,
	HCI_UP,
	HCI_RUNNING,
};

#define BLUEZ_SKB_RESERVE	8

#define HCI_MAX_FRAME_SIZE	2048

/* HCI Packet types */
#define HCI_COMMAND_PKT		0x01
#define HCI_ACLDATA_PKT 	0x02
#define HCI_SCODATA_PKT 	0x03
#define HCI_EVENT_PKT		0x04

/* --------  HCI Packet structures  -------- */
#define HCI_TYPE_LEN		1
#define HCI_COMMAND_HDR_SIZE 	3
#define HCI_EVENT_HDR_SIZE 	2
#define HCI_ACL_HDR_SIZE 	4
#define HCI_SCO_HDR_SIZE 	3

typedef struct {
	__u16 	opcode;		/* OCF & OGF */
	__u8 	plen;
} __attribute__ ((packed))	hci_command_hdr;

typedef struct {
	__u8 	evt;
	__u8 	plen;
} __attribute__ ((packed))	hci_event_hdr;

typedef struct {
	__u16 	handle;		/* Handle & Flags(PB, BC) */
	__u16 	dlen;
} __attribute__ ((packed))	hci_acl_hdr;

typedef struct {
	__u16 	handle;
	__u8 	dlen;
} __attribute__ ((packed))	hci_sco_hdr;


/* ----- SKB helpers ----- */
struct bluez_skb_cb {
	int    incomming;
};
#define bluez_cb(skb)	((struct bluez_skb_cb *)(skb->cb)) 


static inline struct sk_buff *bluez_skb_alloc(unsigned int len, int how)
{
	struct sk_buff *skb;

	if ((skb = alloc_skb(len + BLUEZ_SKB_RESERVE, how))) {
		skb_reserve(skb, BLUEZ_SKB_RESERVE);
		bluez_cb(skb)->incomming  = 0;
	}
	return skb;
}



/* ----- HCI Devices ----- */

struct hci_dev_stats {
	__u32 err_rx;
	__u32 err_tx;
	__u32 cmd_tx;
	__u32 evt_rx;
	__u32 acl_tx;
	__u32 acl_rx;
	__u32 sco_tx;
	__u32 sco_rx;
	__u32 byte_rx;
	__u32 byte_tx;
};

struct hci_dev {
	char			name[8];
	void			*driver_data;
	__u32	 		flags;
	__u8	 		type;
	struct hci_dev_stats 	stat;
	
	int (*open)(struct hci_dev *hdev);
	int (*close)(struct hci_dev *hdev);
	int (*flush)(struct hci_dev *hdev);
	int (*send)(struct sk_buff *skb);

	void (*destruct)(struct hci_dev *hdev);
	int (*ioctl)(struct hci_dev *hdev, unsigned int cmd, unsigned long arg);

	/* Affix */
	struct net_device	*dev;
	struct net_device_stats	net_stats;
};

static inline int hci_dev_init(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (test_and_set_bit(HCI_INIT, &(hdev->flags)))
		return 0;
	DBFEXIT;
	return 0;
}

static inline void hci_dev_uninit(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (!test_and_clear_bit(HCI_INIT, &(hdev->flags)))
		return;
	DBFEXIT;
}

static inline int hci_dev_open(struct net_device *dev)
{
	int		err;
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (test_and_set_bit(HCI_UP, &(hdev->flags)))
		return 0;

	DBPRT(DBINIT, "opening device ...\n");
	err = hdev->open(hdev);
	DBPRT(DBINIT, "device opened: %d\n", err);
	if (err)
		return err;
	hcidev_start_queue(hdev->dev);
	DBFEXIT;
	return 0;
}

static inline int hci_dev_stop(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;
	int		err;

	if (!test_and_clear_bit(HCI_UP, &(hdev->flags)))
		return 0;

	hcidev_stop_queue(hdev->dev);
	err = hdev->close(hdev);
	return err;
}

static inline int hci_dev_ioctl(struct net_device *dev, void *arg, int cmd)
{
	DBFENTER;
	switch (cmd) {
	default:
		return -ENOIOCTLCMD;
	}
	DBFEXIT;
	return 0;
}

static inline struct net_device_stats *hci_dev_get_stats(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	return &hdev->net_stats;
}

static inline int hci_dev_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;
	int		err;

	DBFENTER;
	
	if (!hcidev_running(dev)) {
		dev_kfree_skb(skb);
		return 0;
	}

	if (hcidev_queue_stopped(dev)) {
		return -EBUSY;
	}

	dev->trans_start = jiffies;
	skb->dev = (struct net_device*)hdev;

//	hcidev_stop_queue(dev);
	err = hdev->send(skb);
//	hcidev_wake_queue(hdev->dev);

	DBPRT(DBXMIT, "packet sent to driver: %d\n", err);
	DBFEXIT;
        return err;
}

/* Register HCI device */
static inline int hci_register_dev(struct hci_dev *hdev)
{
	struct net_device	*dev;
	int			err;

	DBFENTER;

	dev = hcidev_alloc();
	if (dev == NULL)
		return -ENOMEM;

	/* bluez stuff */
	memset(&hdev->stat, 0, sizeof(struct hci_dev_stats));
	hdev->flags = 0;
	hdev->dev = dev;

	/* Affix stuff */
	dev->priv = hdev;	/* set private pointer */
	dev->init = hci_dev_init;
	dev->uninit = hci_dev_uninit;
	dev->open = hci_dev_open;
	dev->stop = hci_dev_stop;
	dev->do_ioctl = (void*)hci_dev_ioctl;
	dev->get_stats = hci_dev_get_stats;
	dev->hard_start_xmit = hci_dev_xmit;
	dev->hard_header_len = BLUEZ_SKB_RESERVE;
	dev->if_port = hdev->type;

	SET_MODULE_OWNER(dev);

	err = hcidev_register(dev, NULL);

	DBFEXIT;

	return (err)? err : dev->ifindex;
}

/* Unregister HCI device */
static inline int hci_unregister_dev(struct hci_dev *hdev)
{
	DBFENTER;
	hcidev_unregister(hdev->dev);
	hdev->dev = NULL;
	DBFEXIT;
	return 0;
}

/* Receive frame from HCI drivers */
static inline int hci_recv_frame(struct sk_buff *skb)
{
	struct hci_dev *hdev = (struct hci_dev *) skb->dev;

	DBFENTER;
	if (!hdev || !(hdev->flags & (HCI_UP | HCI_INIT))) {
		kfree_skb(skb);
		return -1;
	}
	skb->dev = hdev->dev;
	hcidev_rx(skb);	/* send to upper protocol layer */
	DBFEXIT;
	return 0;
}

#endif

