/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/* 
   $Id: rfcomm.c,v 1.55 2002/12/13 19:51:56 kds Exp $

   RFCOMM - RFCOMM protocol for Bluetooth

	Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
		Imre Deak <ext-imre.deak@nokia.com>
*/		

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>

#include <linux/interrupt.h>
#include <linux/in.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/io.h>

#include <linux/skbuff.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <net/sock.h>

#define FILEBIT	DBRFCOMM

/* Local Includes */
#include <affix/bluetooth.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/l2cap.h>

#include <affix/rfcomm.h>
//#include "af_rfcomm.h"
#include "bty.h"


/* Local Variables */

btlist_head_t	rfcomm_protos;	/* upper protocol list */
btlist_head_t	rfcomm_sns;	/* sessions list */
__u8 		crctable[256];

/* references */
extern l2cap_proto_ops rfcomm_ops;

/* Client protocol management */
static int	channel_mask = 0;

/* 1.1. spec requires it */
int		credit_based = 1;


/*
 * rfcomm client protocol stuff
 */
rfcomm_proto *__rfcomm_proto_lookup(__u8 channel)
{
	rfcomm_proto	*proto;

	btl_for_each (proto, rfcomm_protos)
		if (proto->channel == channel)
			break;
	return proto;
}

int rfcomm_register_protocol(__u16 *channel, rfcomm_proto_ops *ops, void *param)
{
	int		err = -EBUSY, mask = 1, sch = *channel;
	rfcomm_proto	*proto;

	DBFENTER;

	btl_write_lock(&rfcomm_protos);

	/* check if channel is static */
	if (sch != 0) {
		mask = 1 << (sch-1);
		if ((channel_mask & mask) != 0) {
			err = -EBUSY;
			goto exit;
		}
	} else {
		/* 
		   find free server channel. 
		   channel_mask is guarded with rfcomm_protos write lock
		   */
		for (sch = 1, mask = 1;  sch <= 30; sch++, mask <<= 1)
			if ((channel_mask & mask) == 0)
				break;
	}
	if (sch <= 30) {
		proto = kmalloc(sizeof(*proto), GFP_ATOMIC);
		if (proto == NULL) {
			err = -ENOMEM;
			goto exit;
		}

		channel_mask |= mask;
		*channel = sch;

		proto->channel = sch;
		proto->ops = ops;
		__btl_add_tail(&rfcomm_protos, proto);
		err = 0;
	}

exit:
	btl_write_unlock(&rfcomm_protos);

	DBFEXIT;
	return err;
}

void rfcomm_unregister_protocol(__u16 channel)
{
	rfcomm_proto	*proto;

	DBFENTER;

	btl_write_lock(&rfcomm_protos);
	proto = __rfcomm_proto_lookup(channel);
	if (proto != NULL) {
		__btl_unlink(&rfcomm_protos, proto);
		kfree(proto);
		/*  note that channel mask is guarded with rfcomm_protos write lock. */
		channel_mask &= ~(1<<(channel-1));
	}
	btl_write_unlock(&rfcomm_protos);
	DBFEXIT;
}

/*
 * connection stuff
 */
rfcomm_con *__rfcon_lookup(rfcomm_sn *sn, __u8 dlci)
{
	rfcomm_con	*con;

	DBFENTER;

	btl_for_each (con, sn->cons)
		if (STATE(con) != DEAD && con->dlci == dlci)
			break;

	DBFEXIT;
	return con;
}

rfcomm_con *rfcon_lookup(rfcomm_sn *sn, __u8 dlci)
{
	rfcomm_con	*con;

	DBFENTER;

	btl_read_lock(&sn->cons);
	con = __rfcon_lookup(sn, dlci);
	if (con)
		rfcon_hold(con);
	btl_read_unlock(&sn->cons);

	DBFEXIT;
	return con;
}



rfcomm_con *rfcon_create(BD_ADDR *bda, __u8 dlci, rfcomm_proto_ops *ops)
{
	rfcomm_con	*con;

	DBFENTER;

	con = kmalloc(sizeof(rfcomm_con), GFP_ATOMIC);
	if (con == NULL)
		return NULL;
	memset(con, 0, sizeof(rfcomm_con));

	/* initialize some members */
	init_timer(&con->timer);
	con->timer.data = (unsigned long)con;
	con->timer.function = rfcon_timer;

	con->line = -1;
	con->ops = ops;
	atomic_set(&con->refcnt, 1);	/* use it */
	con->bda = *bda;
	con->dlci = dlci & 0x3F;

	/* Set protocol default parameters */
	/* modem */
	con->peer_modem.ea = 1;
	con->peer_modem.fc = 1;
	con->peer_modem.mr = 0;
	con->modem.ea = 1;
	con->modem.fc = 1;
	con->modem.mr = 0;

	/* parameters */
	rfcon_setframesize(con, RFCOMM_GUESS_PAYLOAD_SIZE);
	con->peer_frame_size = RFCOMM_GUESS_PAYLOAD_SIZE;

	/* port */
	con->param.bit_rate = 0x08; // 230400;
	con->param.data_bits = RFCOMM_CS8;
	con->param.stop_bit = 1; // 1.5 stop bits;
	con->param.parity = 0;
	con->param.parity_type = RFCOMM_PARODD;

	/* credits */
	con->overall_credit = 0;	/* do not accept anything */

	ENTERSTATE(con, CON_CLOSED);	/* set initial state */

	DBFEXIT;
	return con;
}

void __rfcon_destroy(rfcomm_con *con)
{
	DBFENTER;

	if (atomic_read(&con->refcnt) == 0) {
		if (con->sn) {
			__btl_unlink(&con->sn->cons, con);
			rfcomm_put(con->sn);
		}
		if (con->hci)
			hci_put(con->hci);
		kfree(con);
	}

	DBFEXIT;
}

void rfcon_destroy(rfcomm_con *con)
{
	rfcomm_sn	*sn = con->sn;

	DBFENTER;

	/* preserve session from freeing to use locks */
	if (sn) {
		rfcomm_hold(sn);
		btl_write_lock(&sn->cons);
	}
	__rfcon_destroy(con);
	if (sn) {
		btl_write_unlock(&sn->cons);
		rfcomm_put(sn);
	}

	DBFEXIT;
}

static inline int rfcon_unused(rfcomm_con *con)
{
	if (atomic_read(&con->refcnt) == 1 || con->priv == NULL)
		return 1;
	return 0;
}


void rfcon_put(rfcomm_con *con)
{
	DBFENTER;
	
	if (rfcon_unused(con))
		rfcon_disconnect_req(con);
	
	if (atomic_dec_and_test(&con->refcnt))
		rfcon_destroy(con);

	DBFEXIT;
}


int rfcon_connect_req(rfcomm_con *con)
{
	rfcomm_sn	*sn;
	int		err = 0;
	rfcomm_con	*c;

	DBFENTER;

	err = rfcomm_connect_req(con, &sn);
	if (err) {
		DBPRT(DBXMIT, "Unable to create RFCOMM session\n");
		return err;
	}

	con->dlci |= (!sn->initiator & 0x01);	/* set session bit */

	/* check here if session already has connection */
	c = rfcon_lookup(sn, con->dlci);
	if (c) {
		rfcon_put(c);
		rfcomm_put(sn);
		return -EEXIST;
	}

	ENTERSTATE(con, CON_CLOSED);
	rfcon_bind_session(con, sn);
	rfcomm_put(sn);

	if (!rfcomm_connected(sn))
		return 0;

	con->hdrlen = sn->hdrlen + LONG_HDR_SIZE + 1;

	if (STATE(con) == CON_CLOSED) {
		ENTERSTATE(con, CON_W4_CONRSP);
		//con->dlci |= (!sn->initiator & 0x01);	/* XXX: check. set session bit */
		send_sabm(sn, con->dlci);
		rfcon_start_timer(con, RFCOMM_CONRSP_TIMEOUT);	/* T1_TIMEOUT */
	}

	DBFEXIT;
	return err;
}

int rfcon_disconnect_req(rfcomm_con *con)
{
	rfcomm_sn	*sn = con->sn;

	DBFENTER;

	if (STATE(con) == CON_OPEN) {
		ENTERSTATE(con, CON_W4_DISCRSP);
		send_disc(sn, con->dlci);
		rfcon_start_timer(con, RFCOMM_DISCRSP_TIMEOUT);
	} else if (STATE(con) == CON_W4_CONRSP) {
		_send_dm(sn, con->dlci, 0);	// unsolicited
		rfcon_start_timer(con, RFCOMM_DISCRSP_TIMEOUT);
	}

	DBFEXIT;
	return 0;
}

int rfcon_connect_rsp(rfcomm_con *con, int status)
{
	rfcomm_sn	*sn = con->sn;
	int		err = 0;

	DBFENTER;
	
	if (STATE(con) == CON_W4_LCONRSP) {
		if (status == 0) {
			if (con->sn && con->sn->credit_based) {
				con->credit_based = con->sn->credit_based;
				con->local_credit = con->sn->local_credit;
				con->remote_credit = con->sn->remote_credit;
			}
			ENTERSTATE(con, CON_OPEN);
			send_ua(sn, con->dlci);
			send_cmd_msc(con, MCC_CMD);
		} else {
			ENTERSTATE(con, CON_CLOSED);
			send_dm(sn, con->dlci);
		}
	}

	DBFEXIT;

	return err;
}

/* to socket level */

static inline int rfcon_control_ind(rfcomm_con *con, int type)
{
	DBFENTER;

	if (con->priv)
		con->ops->control_ind(con, type);

	DBFEXIT;
	return 0;
}

static inline int rfcon_connect_ind(rfcomm_con *con)
{
	DBFENTER;

	con->ops->connect_ind(con);

	DBFEXIT;
	return 0;
}

int rfcon_connect_cfm(rfcomm_con *con, int status)
{
	DBFENTER;

	if (status == 0) {
		if (con->sn && con->sn->credit_based) {
			con->credit_based = con->sn->credit_based;
			con->local_credit = con->sn->local_credit;
			con->remote_credit = con->sn->remote_credit;
		}
	}
	if (con->priv)
		con->ops->connect_cfm(con, status);

	DBFEXIT;
	return 0;
}

static inline int rfcon_disconnect_ind(rfcomm_con *con)
{
	DBFENTER;

	if (con->priv)
		con->ops->disconnect_ind(con);

	DBFEXIT;
	return 0;
}

/*
 * connection timer
 */
void rfcon_timer(unsigned long data)
{
	rfcomm_con	*con = (rfcomm_con*)data;

	DBFENTER;

	switch (SETSTATE(con, DEAD)) {
		case CON_W4_CONRSP:
			rfcon_connect_cfm(con, STATUS_FAILURE);
			break;
		case CON_CLOSED:
			if (SUBSTATE(con) == RFCON_CFG_PN) {
				if (STATE(con->sn) == CON_CONFIG)
					rfcomm_config_complete(con->sn, 0);
			}
			break;
		case CON_W4_CONREQ:
			con->priv = NULL;
			break;
		default:
	}
	rfcon_put(con);

	DBFEXIT;
}


/*
 * session management
 */

static inline rfcomm_sn *__rfcomm_lookup(BD_ADDR *bda)
{
	rfcomm_sn	*sn;

	DBFENTER;

	btl_for_each (sn, rfcomm_sns)
		if (STATE(sn) != DEAD && memcmp(&sn->bda, bda, 6) == 0)
			break;

	DBFEXIT;
	return sn;
}

rfcomm_sn *rfcomm_create(BD_ADDR *bda)
{
	rfcomm_sn	*sn;

	DBFENTER;

	btl_write_lock(&rfcomm_sns);
	sn = __rfcomm_lookup(bda);
	if (sn == NULL) {
		sn = kmalloc(sizeof(rfcomm_sn), GFP_ATOMIC);
		if (sn == NULL) {
			BTERROR("Memory allocation failed\n");
			goto exit;
		}
		memset(sn, 0, sizeof(rfcomm_sn));

		/* initialize internal data structure */
		init_timer(&sn->timer);
		sn->timer.data = (unsigned long)sn;
		sn->timer.function = rfcomm_timer;

		btl_head_init(&sn->cons);
		memcpy(&sn->bda, bda, 6);
		
		/* set protocol default parameters */
		sn->peer_fc = 1;		/* session is active by default */
		sn->fc = 1;			/* disabled transmission */
		sn->credit_based = 0;		/* not credit based */
		sn->local_credit = 0; 		/* disabled transmission, credits - 0 */

		ENTERSTATE(sn, CON_CLOSED);	/* initial state */
		__btl_add_tail(&rfcomm_sns, sn);
	}
	rfcomm_hold(sn);
exit:
	btl_write_unlock(&rfcomm_sns);
	DBFEXIT;
	return sn;
}

int rfcomm_destroy(rfcomm_sn *sn)
{
	DBFENTER;
	if (atomic_read(&sn->refcnt) == 0) {
		btl_unlink(&rfcomm_sns, sn);
		if (sn->ch) {
			sn->ch->priv = NULL;
			l2ca_put(sn->ch);
		}
		kfree(sn);
	}
	DBFEXIT;
	return 0;
}

rfcomm_con *find_active(btlist_head_t *cons)
{
	rfcomm_con	*con;

	btl_read_lock(cons);
	btl_for_each (con, *cons) {
		if (STATE(con) != DEAD) {
			rfcon_hold(con);
			break;
		}
	}
	btl_read_unlock(cons);

	return con;
}

static inline int rfcomm_unused(rfcomm_sn *sn)
{
	if (atomic_read(&sn->refcnt) == 1) {
		rfcomm_con	*con;

		con = find_active(&sn->cons);
		if (con)
			rfcon_put(con);
		else
			return 1;
	}
	return 0;
}

void rfcomm_put(rfcomm_sn *sn)
{
	DBFENTER;
	
	if (STATE(sn) == CON_OPEN && rfcomm_unused(sn)) {
		/* start defer disconnection timer */
		rfcomm_start_timer(sn, RFCOMM_DEFER_DISC_TIMEOUT);
	}
	if (atomic_dec_and_test(&sn->refcnt))
		rfcomm_destroy(sn);

	DBFEXIT;
}

void rfcomm_timer(unsigned long data)
{
	rfcomm_sn	*sn = (rfcomm_sn*)data;

	DBFENTER;

	switch (STATE(sn)) {
		case CON_W4_CONRSP:
			ENTERSTATE(sn, DEAD);
			rfcomm_connect_cfm(sn, STATUS_FAILURE);
			break;
		case CON_OPEN:
			/* defer disc timeout, ... */
			if (rfcomm_unused(sn))
				rfcomm_disconnect_req(sn);
			break;
		default:
			ENTERSTATE(sn, DEAD);
	}
	rfcomm_put(sn);

	DBFEXIT;
}


/* 
   1. Create and configure L2CAP channel
   2. Send SABM command on DLCI 0
   */
static inline int __rfcomm_connect_req(hci_struct *hci, rfcomm_sn *sn)
{
	l2cap_ch	*ch;
	int		err;

	DBFENTER;

	sn->initiator = 1;
	ch = l2ca_create(PSM_RFCOMM, &rfcomm_ops);
	if (ch == NULL) {
		err = -ENOMEM;
		goto exit;
	}
	sn->ch = ch;
	ch->priv = sn;
	if (hci)
		hci_hold(hci);	// set reference
	ch->hci = hci;
	l2ca_setmtu(ch, LONG_HDR_SIZE + RFCOMM_GUESS_PAYLOAD_SIZE);
	err = l2ca_connect_req(ch, &sn->bda);
	sn->hdrlen = ch->hdrlen;
exit:
	DBFEXIT;
	return err;
}

int rfcomm_connect_req(rfcomm_con *con, rfcomm_sn **ret)
{
	int		err = 0;
	rfcomm_sn	*sn;

	sn = rfcomm_create(&con->bda);
	if (sn == NULL)
		return -ENOMEM;

	if (STATE(sn) == CON_CLOSED) {
		ENTERSTATE(sn, CON_W4_CONRSP);
		err = __rfcomm_connect_req(con->hci, sn);
	}

	*ret = sn;

	DBFENTER;
	return err;
}

/*
 * session is ready to use
 */

int rfcomm_config_complete(rfcomm_sn *sn, int status)
{
	rfcomm_con	*con, *next;

	DBFENTER;
	
	btl_write_lock(&sn->cons);
	btl_for_each_next (con, sn->cons, next) {
		rfcon_hold(con);
		if (status == 0) {
			con->hdrlen = sn->hdrlen + LONG_HDR_SIZE + 1;
			if (STATE(con) == CON_CLOSED) {
				ENTERSTATE(con, CON_W4_CONRSP);
				con->dlci |= (!sn->initiator & 0x01);	/* session bit */
				send_sabm(sn, con->dlci);		/* Set SABM */
				rfcon_start_timer(con, RFCOMM_CONRSP_TIMEOUT); /* T1_TIMEOUT) */
			}
		} else {
			ENTERSTATE(con, DEAD);
			rfcon_connect_cfm(con, STATUS_FAILURE);
		}
		__rfcon_put(con);
	}
	btl_write_unlock(&sn->cons);

	DBFEXIT;
	return 0;
}

		
int rfcomm_connect_cfm(rfcomm_sn *sn, int status)
{
	rfcomm_con	*con;
	int		err = 0;

	DBFENTER; 

	if (status)
		return rfcomm_config_complete(sn, status);

	if (!credit_based) {
		ENTERSTATE(sn, CON_OPEN);
		return rfcomm_config_complete(sn, status);
	}

	if (sn->initiator) {
		/* session initiator */
		sn->credit_based = 1;	/* try to enable it */
		/* take first not dead */
		con = find_active(&sn->cons);
		if (con) {
			con->hdrlen = sn->hdrlen + LONG_HDR_SIZE + 1;

			/* FIXME: check state here */
			con->credit_based = sn->credit_based;
			con->local_credit = sn->local_credit;
			/* session in CONFIG state */
			ENTERSUBSTATE(con, RFCON_CFG_PN);
			send_cmd_pn(con, MCC_CMD);
			rfcon_start_timer(con, T2_TIMEOUT);
			rfcon_put(con);
		}
	} else {
		/* not initator */
		rfcomm_start_timer(sn, RFCOMM_CFGREQ_TIMEOUT);
	}
	
	DBFEXIT;
	return err;
}

int rfcomm_disconnect_ind(rfcomm_sn *sn)
{
	rfcomm_con	*con, *next;

	DBFENTER;

	btl_write_lock(&sn->cons);
	btl_for_each_next (con, sn->cons, next) {
		rfcon_hold(con);
		switch (SETSTATE(con, DEAD)) {
			case CON_OPEN:
			case CON_W4_LCONRSP:
				rfcon_disconnect_ind(con);
				break;
			case CON_CLOSED:
				/* wait for link up */
			case CON_W4_CONRSP:
				rfcon_stop_timer(con);
				rfcon_connect_cfm(con, STATUS_FAILURE);
				break;
			default:
		}
		__rfcon_put(con);
	}
	btl_write_unlock(&sn->cons);

	DBFEXIT;
	return 0;
}

int rfcomm_disconnect_req(rfcomm_sn *sn)
{
	DBFENTER;

	if (STATE(sn) == CON_OPEN || STATE(sn) == CON_CONFIG) {
		ENTERSTATE(sn, CON_W4_DISCRSP);
		send_disc(sn, 0);
		rfcomm_start_timer(sn, RFCOMM_DISCRSP_TIMEOUT);
		return 1;
	}

	DBFEXIT;
	return 0;
}


/* 
 * RFCOMM transmitting stuff
 */
static inline int rfcomm_send(rfcomm_sn *sn, struct sk_buff *skb)
{
	int	err;

	DBFENTER;
	DBPARSERFCOMM(DBXMIT, skb->data, skb->len, TO_HOSTCTRL);
	DBDUMPCHAR(DBXMIT, skb->data, skb->len);
	DBPRTPKT(DBXMIT, skb->data, skb->len);

	err = l2ca_send_data(sn->ch, skb);
	if (err)
		kfree_skb(skb);

	return err;
}

int send_sabm(rfcomm_sn *sn, __u8 dlci)
{
	struct sk_buff	*skb;
	short_frame	*pkt;

	DBFENTER;

	BTASSERT(sn, return 0;);

	DBPRT(DBCTRL, "Creating SABM packet to DLCI %d, hdrlen: %d\n", dlci, sn->hdrlen);

	skb = alloc_skb(sn->hdrlen+SABM_SIZE, GFP_ATOMIC);
	if (skb == NULL) return -ENOMEM;
	skb_reserve(skb, sn->hdrlen);
	skb_put(skb, SABM_SIZE);

	pkt = (short_frame*)skb->data;
	pkt->addr.ea = 1;		/* only one address field */
	pkt->addr.cr = sn->initiator;
	pkt->addr.dlci = dlci;
	pkt->control = SET_PF(SABM);
	pkt->length.ea = 1;
	pkt->length.len = 0;
	pkt->data[0] = crc_calc((__u8*) pkt, CTRL_CRC_CHECK);

	return rfcomm_send(sn, skb);
}

int send_ua(rfcomm_sn *sn, __u8 dlci)
{
	struct sk_buff	*skb;
	short_frame	*pkt;

	DBFENTER;

	BTASSERT(sn, return 0;);

	DBPRT(DBCTRL, "Creating UA packet to DLCI %d\n",  dlci);

	skb = alloc_skb(sn->hdrlen+UA_SIZE, GFP_ATOMIC);	/* +4 */
	if (skb == NULL) return -ENOMEM;
	skb_reserve(skb, sn->hdrlen);
	skb_put(skb, UA_SIZE);

	pkt = (short_frame*)skb->data;
	pkt->addr.ea = 1;		/* only one address field */
	pkt->addr.cr = !sn->initiator;
	pkt->addr.dlci = dlci;
	pkt->control = SET_PF(UA);
	pkt->length.ea = 1;
	pkt->length.len = 0;
	pkt->data[0] = crc_calc((__u8*) pkt, CTRL_CRC_CHECK);

	return rfcomm_send(sn, skb);
}

int _send_dm(rfcomm_sn *sn, __u8 dlci, int solicited)
{
	struct sk_buff	*skb;
	short_frame	*pkt;

	DBFENTER;

	BTASSERT(sn, return 0;);

	DBPRT(DBCTRL, "Creating DM packet to DLCI %d\n", dlci);

	skb = alloc_skb(sn->hdrlen+UA_SIZE, GFP_ATOMIC);
	if (skb == NULL) return -ENOMEM;
	skb_reserve(skb, sn->hdrlen);
	skb_put(skb, UA_SIZE);

	pkt = (short_frame*)skb->data;
	pkt->addr.ea = 1;		/* only one address field */
	pkt->addr.cr = !sn->initiator;
	pkt->addr.dlci = dlci;
	pkt->control = solicited?SET_PF(DM):CLR_PF(DM);
	pkt->length.ea = 1;
	pkt->length.len = 0;
	pkt->data[0] = crc_calc((__u8*) pkt, CTRL_CRC_CHECK);

	return rfcomm_send(sn, skb);
}

int send_dm(rfcomm_sn *sn, __u8 dlci)
{
	return _send_dm(sn, dlci, 1);
}

int send_disc(rfcomm_sn *sn, __u8 dlci)
{
	struct sk_buff	*skb;
	short_frame	*pkt;

	DBFENTER;

	BTASSERT(sn, return 0;);

	DBPRT(DBCTRL, "Creating DISC packet to DLCI %d\n", dlci);

	skb = alloc_skb(sn->hdrlen+UA_SIZE, GFP_ATOMIC);
	if (skb == NULL) return -ENOMEM;
	skb_reserve(skb, sn->hdrlen);
	skb_put(skb, UA_SIZE);

	pkt = (short_frame*)skb->data;
	pkt->addr.ea = 1;		/* only one address field */
	pkt->addr.cr = sn->initiator;
	pkt->addr.dlci = dlci;
	pkt->control = SET_PF(DISC);
	pkt->length.ea = 1;
	pkt->length.len = 0;
	pkt->data[0] = crc_calc((__u8*) pkt, CTRL_CRC_CHECK);

	return rfcomm_send(sn, skb);
}

void set_uih_hdr(short_frame *pkt, __u8 dlci, uint len, __u8 cr)
{
	DBFENTER;

	pkt->addr.ea = 1;
	pkt->addr.cr = cr;
	pkt->addr.dlci = dlci;
	pkt->control = CLR_PF(UIH);

	if(len > SHORT_PAYLOAD_SIZE) {
		((long_frame*) pkt)->length.len = SET_LENGTH(len);  
	} else {
		pkt->length.ea = 1;
		pkt->length.len = len;  
	}

	DBFEXIT;
}

int rfcomm_send_credit(rfcomm_con *con, int credit)
{
	rfcomm_sn	*sn = con->sn;
	struct sk_buff	*skb;
	credit_short_frame	*pkt;

	DBFENTER;

	if (credit <= 0)
		return 0;

	if (credit > 0xFF)
		credit = 0xFF;	/* we send up to byte */

	DBPRT(DBXMIT, "Sending credits: %d\n", credit);
	skb = alloc_skb(sn->hdrlen+sizeof(*pkt)+FCS_SIZE, GFP_ATOMIC);
	if (skb == NULL)
		return -ENOMEM;
	skb_reserve(skb, sn->hdrlen);
	pkt = (credit_short_frame*)skb_put(skb, sizeof(*pkt));

	set_uih_hdr((void*)pkt, con->dlci, 0, sn->initiator);
	pkt->control = SET_PF(UIH);
	pkt->credit = credit;
	con->local_credit += credit;
	DBPRT(DBRECV, "sCREDIT1x: Local: %d, Overall: %d, Remote: %d\n",
			con->local_credit, con->overall_credit, con->remote_credit);
	*skb->tail = crc_calc((__u8*) pkt, UIH_CRC_CHECK);
	skb_put(skb, FCS_SIZE);		/* add FCS */

	DBFEXIT;
	return rfcomm_send(sn, skb);
}

int rfcomm_send_data(rfcomm_con *con, struct sk_buff *skb)
{
	rfcomm_sn	*sn = con->sn;
	int		flag = 0, len = skb->len;	/* data length */
	__u8		*credit;
	short_frame	*pkt;

	DBFENTER;

	DBPRT(DBCTRL, "Creating UIH packet, len: %d, dlci: %d, headroom: %d\n",
			len, con->dlci, skb_headroom(skb));

	if (rfcon_disabled(con)) {/* we can't send now */
		DBPRT(DBXMIT, "Connection disabled\n");
		kfree_skb(skb);
		return -EBUSY;
	}

	if (con->credit_based) {
		DBPRT(DBRECV, "sCREDIT1: Local: %d, Overall: %d, Remote: %d\n",
				con->local_credit, con->overall_credit, con->remote_credit);
		con->remote_credit--;	/* decrease remote credit */
		if (con->remote_credit == 0)
			rfcon_control_ind(con, RFCOMM_TX_STOP);

		if ((con->local_credit << 1) <= con->overall_credit) {
			int	diff;

			rfcon_control_ind(con, RFCOMM_BUFFER_REQUEST);
			diff = con->overall_credit - con->local_credit;
			if (diff > 0) {
				flag = 1;
				credit = skb_push(skb, 1);
				*credit = diff;
				con->local_credit += *credit;
			}
		}
		DBPRT(DBRECV, "sCREDIT2: Local: %d, Overall: %d, Remote: %d\n",
				con->local_credit, con->overall_credit, con->remote_credit);
	}

	if (len > SHORT_PAYLOAD_SIZE)
		pkt = (short_frame*)skb_push(skb, LONG_HDR_SIZE);
	else
		pkt = (short_frame*)skb_push(skb, SHORT_HDR_SIZE);


	set_uih_hdr(pkt, con->dlci, len, sn->initiator);
	if (flag)
		pkt->control = SET_PF(UIH);
	*skb->tail = crc_calc((__u8*) pkt, UIH_CRC_CHECK);
	skb_put(skb, FCS_SIZE);		/* add FCS */

	return rfcomm_send(sn, skb);
}	

/*
   assume that skb contains pure command
   */
int rfcomm_send_cmd(rfcomm_sn *sn, struct sk_buff *skb, __u8 type, __u8 cr)
{
	short_frame		*pkt;
	mcc_frame		*cmd;
	int			cmdlen = skb->len;
	int			len;

	DBFENTER;

	cmd = (mcc_frame*)skb_push(skb, sizeof(mcc_frame));
	cmd->type.ea = 1;
	cmd->type.cr = cr;
	cmd->type.value = type;
	cmd->length.ea = 1;
	cmd->length.len = cmdlen;

	len = skb->len;
	if (len > SHORT_PAYLOAD_SIZE)
		pkt = (short_frame*)skb_push(skb, LONG_HDR_SIZE);
	else
		pkt = (short_frame*)skb_push(skb, SHORT_HDR_SIZE);

	set_uih_hdr(pkt, 0, len, sn->initiator);
	*skb->tail = crc_calc((__u8*) pkt, UIH_CRC_CHECK);
	skb_put(skb, FCS_SIZE);		/* add FCS */

	return rfcomm_send(sn, skb);
}	


int send_cmd_msc(rfcomm_con *con, __u8 cr)
{
	rfcomm_sn	*sn = con->sn;
	msc_msg		*msg;
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, sizeof(*msg));
	msg = (msc_msg*)skb->data;
	/* set fields */
	msg->addr.ea = 1;
	msg->addr.cr = 1;
	msg->addr.dlci = con->dlci;
	msg->sig = (cr)?con->modem:con->peer_modem;

	/* send */
	rfcomm_send_cmd(sn, skb, R_MSC, cr);

	DBFEXIT;
	return 0;
}

int send_cmd_rpn(rfcomm_con *con, __u8 cr)
{
	rfcomm_sn	*sn = con->sn;
	rpn_msg		*msg;
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, sizeof(*msg));
	msg = (rpn_msg*)skb->data;
	/* set fields */
	msg->addr.ea = 1;
	msg->addr.cr = 1;
	msg->addr.dlci = con->dlci;
	msg->param = (cr)?con->peer_param:con->param;
	msg->pm = (cr)?con->peer_pm:con->pm;

	/* send */
	rfcomm_send_cmd(sn, skb, RPN, cr);

	DBFEXIT;
	return 0;
}

int send_cmd_rls(rfcomm_con *con, __u8 cr)
{
	rfcomm_sn	*sn = con->sn;
	rls_msg		*msg;
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, sizeof(*msg));
	msg = (rls_msg*)skb->data;
	/* set fields */
	msg->addr.ea = 1;
	msg->addr.cr = 1;
	msg->addr.dlci = con->dlci;
	msg->line_status = (cr)?con->line_status:con->peer_line_status;

	/* send */
	rfcomm_send_cmd(sn, skb, RLS, cr);

	DBFEXIT;
	return 0;
}

int send_cmd_pn(rfcomm_con *con, __u8 cr)
{
	rfcomm_sn	*sn = con->sn;
	pn_msg		*msg;
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, sizeof(*msg));
	msg = (pn_msg*)skb->data;

	/* set fields */
	memset(msg, 0, sizeof(*msg));	/* zero it */
	msg->dlci = con->dlci;
	msg->frame_size = __htob16(con->frame_size);	/* may propose less */
	msg->priority = con->priority;
	if (con->credit_based) {
		msg->flow = (cr)?0x0F:0x0E;
		if (con->local_credit > 0x07)
			con->local_credit = 0x07;
		msg->credit = con->local_credit;	/* let's accept */
	}
	/* send */
	rfcomm_send_cmd(sn, skb, PN, cr);

	DBFEXIT;
	return 0;
}

int send_cmd_nsc(rfcomm_sn *sn, __u8 type, int cr)
{
	nsc_msg		*msg;
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, sizeof(*msg));
	msg = (nsc_msg*)skb->data;
	/* set fields */
	msg->cmd_type.ea = 1;
	msg->cmd_type.cr = cr;
	msg->cmd_type.value = type;

	/* send */
	rfcomm_send_cmd(sn, skb, NSC, MCC_CMD);

	DBFEXIT;
	return 0;
}

int send_cmd_fc(rfcomm_sn *sn, __u8 fc, int cr)
{
	struct sk_buff	*skb;

	DBFENTER;

	skb = rfcomm_alloc_cmd_skb(sn, 0);
	if (fc)
		rfcomm_send_cmd(sn, skb, FCON, cr /* MCC_CMD */);
	else
		rfcomm_send_cmd(sn, skb, FCOFF, cr /* MCC_CMD */);

	DBFEXIT;
	return 0;
}


/* receiving side */

/* packets processing */

int recv_sabm(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	rfcomm_con	*con;
	int		err = 0;
	__u8		dlci = pkt->addr.dlci;

	DBFENTER;

	DBPRT(DBRECV, "SABM received, dlci: %d\n", dlci);

	if (dlci == 0) {
		switch (STATE(sn)) {
			case CON_W4_CONREQ:
				rfcomm_stop_timer(sn);	/* conn_timeout */
				ENTERSTATE(sn, CON_CONFIG);
				send_ua(sn, 0);
				rfcomm_connect_cfm(sn, STATUS_SUCCESS);
				break;
			default:
				send_dm(sn, 0);
		}
	} else {
		rfcomm_proto	*proto;

		/* lookup channel */
		btl_read_lock(&rfcomm_protos);
		proto = __rfcomm_proto_lookup(SERVER_CHANNEL(dlci));
		if (proto == NULL) {
			send_dm(sn, dlci);
			goto exit;
		}
		/* check if dlci is already connected */
		con = rfcon_lookup(sn, dlci);
		if (con) {
			if (STATE(con) != CON_W4_CONREQ) {
				/* busy */
				send_dm(sn, dlci);
				rfcon_put(con);
				goto exit;
			}
			/* we got cmd_pn before. stop timer */
			rfcon_stop_timer(con);
			con->priv = NULL;
		} else {
			if (STATE(sn) == CON_CONFIG) {
				/* 
				 * received sabm w/o prior cmd_pn
				 * switch to non credit based mode
				 */
				rfcomm_stop_timer(sn);
				ENTERSTATE(sn, CON_OPEN);
			}
			con = rfcon_create(&sn->bda, dlci, proto->ops);
			if (con == NULL) {
				send_dm(sn, dlci);
				err = -ENOMEM;
				goto exit;
			}
			hci_hold(sn->ch->hci);	// set reference
			con->hci = sn->ch->hci;
			rfcon_bind_session(con, sn);
		}
		ENTERSTATE(con, CON_W4_LCONRSP);
		rfcon_connect_ind(con);
		rfcon_put(con);
exit:
		btl_read_unlock(&rfcomm_protos);
	}
	kfree_skb(skb);
	DBFEXIT;
	return err;
}

int recv_ua(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	rfcomm_con	*con;
	int		err = 0;
	__u8		dlci = pkt->addr.dlci;

	DBFENTER;

	DBPRT(DBRECV, "UA received, dlci: %d\n", dlci);

	if (dlci == 0) {
		rfcomm_stop_timer(sn);
		switch (STATE(sn)) {
			case CON_W4_CONRSP:
				ENTERSTATE(sn, CON_CONFIG);
				rfcomm_connect_cfm(sn, STATUS_SUCCESS);
				break;
			case CON_W4_DISCRSP:
				ENTERSTATE(sn, DEAD);
				break;
			default:
		}
	} else {
		con = rfcon_lookup(sn, dlci);
		if (con  == NULL) {
			DBPRT(DBRECV, "CON not found\n");
			goto exit;
		}
		rfcon_stop_timer(con);
		switch (STATE(con)) {
			case CON_W4_CONRSP:
				ENTERSTATE(con, CON_OPEN);
				send_cmd_msc(con, MCC_CMD);
				rfcon_connect_cfm(con, STATUS_SUCCESS);
				break;

			case CON_W4_DISCRSP:
				ENTERSTATE(con, DEAD);
				break;
			default:
		}
		rfcon_put(con);
	}
exit:
	kfree_skb(skb);
	DBFEXIT;
	return err;
}

int recv_dm(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	rfcomm_con	*con;
	__u8		dlci = pkt->addr.dlci;
	int		err = 0;

	DBFENTER;

	DBPRT(DBRECV, "DM received, dlci: %d\n", dlci);

	if (dlci == 0) {
		rfcomm_stop_timer(sn);
		switch (SETSTATE(sn, DEAD)) {
			case CON_W4_CONRSP:
				rfcomm_connect_cfm(sn, STATUS_FAILURE);
				break;
			case CON_W4_DISCRSP:
				break;
			default:
		}
	} else {
		/* search connecting cons */
		con = rfcon_lookup(sn, dlci);
		if (con  == NULL) {
			DBPRT(DBRECV, "CON not found\n");
			goto exit;
		}
		rfcon_stop_timer(con);
		switch (SETSTATE(con, DEAD)) {
			case CON_CLOSED:
			case CON_W4_CONRSP:
				rfcon_connect_cfm(con, STATUS_FAILURE);
				break;
			case CON_W4_DISCRSP:
				break;
			default:
		}
		rfcon_put(con);
	}
exit:
	kfree_skb(skb);
	DBFEXIT;
	return err;
}

int recv_disc(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	rfcomm_con	*con;
	int		err = 0;
	__u8		dlci = pkt->addr.dlci;

	DBFENTER;

	DBPRT(DBRECV, "DISC received, dlci: %d\n", dlci);

	if (dlci == 0) {
		switch (SETSTATE(sn, DEAD)) {
			default:
				rfcomm_disconnect_ind(sn);
				send_ua(sn, 0);
		}
	} else {
		/* search connecting cons */
		con = rfcon_lookup(sn, dlci);
		if (con == NULL) {
			send_dm(sn, dlci);
			goto exit;
		}

		switch (SETSTATE(con, DEAD)) {
			case CON_W4_DISCRSP:
				/* disc timer is running here
				 * do not ~dead~ con, return to that state
				 */
				ENTERSTATE(con, CON_W4_DISCRSP);
				break;
			case CON_OPEN:
				rfcon_disconnect_ind(con);
				break;
			default:
		}
		send_ua(sn, dlci);
		rfcon_put(con);
	}
exit:
	kfree_skb(skb);
	DBFEXIT;
	return err;
}


int rfcomm_recv_data(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	rfcomm_con	*con;
	int		free = 1;
	__u8		dlci = pkt->addr.dlci, credit;

	DBFENTER;

	if (pkt->length.ea) { /* short */
		skb_pull(skb, SHORT_HDR_SIZE);
	} else {
		skb_pull(skb, LONG_HDR_SIZE);
	}

	DBPRT(DBRECV, "DATA frame has been received, dlci: %d, skblen: %d\n", dlci, skb->len);

	con = rfcon_lookup(sn, dlci);
	if (con == NULL) {
		DBPRT(DBRECV, "CON not found\n");
		kfree_skb(skb);
		return 0;
	}

	if (con->credit_based) {
		DBPRT(DBRECV, "CREDIT1: Local: %d, Overall: %d, Remote: %d\n",
				con->local_credit, con->overall_credit, con->remote_credit);
		if (GET_PF(pkt->control)) { /* read credit */
			credit = *skb->data;
			skb_pull(skb, 1);
			con->remote_credit += credit;
			rfcon_control_ind(con, RFCOMM_TX_START);
		}
		/* check current credit situation */
		if (skb->len != 0 && con->local_credit) {
			con->local_credit--;	/* we can accept one less */
			if (con->priv) {
				con->ops->data_ind(con, skb);
				free = 0;
			}
		}

		DBPRT(DBRECV, "CREDIT2: Local: %d, Overall: %d, Remote: %d\n",
				con->local_credit, con->overall_credit, con->remote_credit);
		rfcon_update_flow(con);
		DBPRT(DBRECV, "CREDIT3: Local: %d, Overall: %d, Remote: %d\n",
				con->local_credit, con->overall_credit, con->remote_credit);
	} else {
		if (skb->len != 0 && con->priv) {
			con->ops->data_ind(con, skb);
			free = 0;
		}
	}
	rfcon_put(con);
	if (free)
		kfree_skb(skb);
	DBFEXIT;
	return 0;
}

/*
 * DLC parameter negotiation
 * should be used before DLC creation
 *
 * response..
 * replies with local proposals
 */

int recv_cmd_pn(rfcomm_sn *sn, struct sk_buff *skb, int cr)
{
	pn_msg		*msg = (pn_msg*)skb->data;
	__u8		dlci = msg->dlci;
	rfcomm_con	*con;

	DBFENTER;

	con = rfcon_lookup(sn, dlci);
	if (cr == MCC_CMD) {
		if (con == NULL) {
			/* PN comes before SABM for DLCI!=0. Spec 1.1 */
			rfcomm_proto	*proto;

			/* lookup channel */
			btl_read_lock(&rfcomm_protos);
			proto = __rfcomm_proto_lookup(SERVER_CHANNEL(dlci));
			if (proto == NULL) {
				send_dm(sn, dlci);
				btl_read_unlock(&rfcomm_protos);
				return -1;
			}
			con = rfcon_create(&sn->bda, dlci, proto->ops);
			if (con == NULL) {
				send_dm(sn, dlci);
				btl_read_unlock(&rfcomm_protos);
				return -ENOMEM;
			}
			hci_hold(sn->ch->hci);	// set reference
			con->hci = sn->ch->hci;
			rfcon_bind_session(con, sn);
			btl_read_unlock(&rfcomm_protos);
			/* this is orphan con now. start timer */
			ENTERSTATE(con, CON_W4_CONREQ);	/* need sabm next */
			con->priv = con;	/* owned by rfcomm entity */
			rfcon_start_timer(con, RFCOMM_CONREQ_TIMEOUT);
		}

		/* store peer parameters */
		con->peer_frame_size = __btoh16(msg->frame_size);
		con->peer_priority = msg->priority;
		/* accept peer frame size as it is */
		rfcon_setframesize(con, con->peer_frame_size);
		con->priority = con->peer_priority;

		/*
		 * credit based stuff
		 */
		if (credit_based && msg->flow == 0x0F) { /* credit based flow request */
			if (STATE(sn) == CON_CONFIG && !sn->credit_based) {
				/* at first PN ... */
				sn->credit_based = 1;
				sn->remote_credit = msg->credit;
			}
		}			
		con->credit_based = sn->credit_based; 
		con->local_credit = sn->local_credit;
		con->remote_credit = msg->credit;

		/* set response local parameters */
		send_cmd_pn(con, MCC_RSP);
		/*
		 * session configured
		 */
		if (STATE(sn) == CON_CONFIG) {
			rfcomm_stop_timer(sn);
			ENTERSTATE(sn, CON_OPEN);
			rfcomm_config_complete(sn, 0);
		}
	} else {
		/* check response here */
		if (con == NULL) {
			send_dm(sn, dlci);
			return -1;
		}

		if (SUBSTATE(con) == RFCON_CFG_PN) {
			rfcon_stop_timer(con);
			ENTERSUBSTATE(con, 0);
		}

		/* store peer parameters */
		con->peer_frame_size = __btoh16(msg->frame_size);
		con->peer_priority = msg->priority;
		/* accept peer frame size as it is */
		rfcon_setframesize(con, con->peer_frame_size);
		con->priority = con->peer_priority;
		/*
		 * check credit based stuff
		 */
		if (STATE(sn) == CON_CONFIG && sn->credit_based) {
			/* check PN response and initiate connections */
			if (msg->flow != 0x0E) {
				/* remote side does not support it */
				sn->credit_based = 0;
			} else
				sn->remote_credit = msg->credit;
		}
		con->credit_based = sn->credit_based; 
		con->local_credit = sn->local_credit;
		con->remote_credit = msg->credit;
		/*
		 * session configured. send sabm for pending connections.
		 */
		if (STATE(sn) == CON_CONFIG) {
			ENTERSTATE(sn, CON_OPEN);
			rfcomm_config_complete(sn, 0);
		}
	}
	rfcon_put(con);

	DBFEXIT;
	return 0;
}

/*
   Remote con line status

   response - received values
   */
int recv_cmd_rls(rfcomm_sn *sn, struct sk_buff *skb, int cr)
{
	rls_msg		*msg = (rls_msg*)skb->data;
	__u8		dlci = msg->addr.dlci;
	rfcomm_con	*con;

	DBFENTER;

	con = rfcon_lookup(sn, dlci);
	if (cr == MCC_CMD) {
		if (con == NULL) {
			send_dm(sn, dlci);
			return -1;
		}
		con->peer_line_status = msg->line_status;
		send_cmd_rls(con, MCC_RSP);
		rfcon_control_ind(con, RFCOMM_LINE_STATUS);
	} else {
		/* response... */
		if (!con)
			return -1;
		/* stop the timer here... */
	}	
	rfcon_put(con);

	DBFEXIT;
	return 0;
}

/*
   Set remote port communication settings

   response - analyze and accept or ... (rules)
   */
int recv_cmd_rpn(rfcomm_sn *sn, struct sk_buff *skb, int cr)
{
	rpn_msg		*msg = (rpn_msg*)skb->data;
	__u8		dlci = msg->addr.dlci;
	rfcomm_con	*con;

	DBFENTER;

	con = rfcon_lookup(sn, dlci);
	if (cr == MCC_CMD) {
		if (con == NULL) {
			send_dm(sn, dlci);
			return -1;
		}
		if (skb->len > sizeof(msg->addr)) {
			/* new settings */
			con->param = msg->param;
			con->pm = msg->pm;
			rfcon_control_ind(con, RFCOMM_PORT_PARAM);
		}
		send_cmd_rpn(con, MCC_RSP);
	} else {
		/* response... */
		if (!con)
			return -1;
		/* stop the timer here... */
	}	
	rfcon_put(con);

	DBFEXIT;
	return 0;
}

/*
   Modem Status Command

   response - received values
   */
int recv_cmd_msc(rfcomm_sn *sn, struct sk_buff *skb, int cr)
{
	mcc_frame	*hdr = (mcc_frame*)(skb->data-sizeof(mcc_frame));
	msc_msg_full	*msg = (msc_msg_full*)skb->data;
	__u8		dlci = msg->addr.dlci;
	rfcomm_con	*con;

	DBFENTER;

	DBPRT(DBCTRL, "MSC cmd length: %d\n", hdr->length.len);

	con = rfcon_lookup(sn, dlci);
	if (cr == MCC_CMD) {
		if (con == NULL) {
			send_dm(sn, dlci);
			return -1;
		}

		con->peer_modem = msg->sig;
		send_cmd_msc(con, MCC_RSP);	/* response */
		rfcon_control_ind(con, RFCOMM_LINE_STATE);
#if 0		
		if (con->bty && con->bty->tty)
			DBPRT(DBCTRL, "HW STOPPED:%d\n", con->bty->tty->hw_stopped);
#endif
		if (hdr->length.len == 3) {
			// we have break signals
		}
	} else {
		/* response... */
		if (!con)
			return -1;
		/* stop the timer here... */
	}
	rfcon_put(con);

	DBFEXIT;
	return 0;
}

int recv_cmd_fc(rfcomm_sn *sn, __u8 fc, int cr)
{
	__u8		delta_flow = sn->peer_fc;
	rfcomm_con	*con;
	__u8		event;

	DBFENTER;

	if (cr == MCC_CMD) {
		sn->peer_fc = fc;
		delta_flow ^= fc;

		if (delta_flow) {
			if (fc == 1)
				event = RFCOMM_TX_START;
			else
				event = RFCOMM_TX_STOP;

			btl_read_lock(&sn->cons);	
			btl_for_each (con, sn->cons)
				rfcon_control_ind(con, event);
			btl_read_unlock(&sn->cons);
		}
		send_cmd_fc(sn, fc, MCC_RSP);
	} else {
		/* response... */
		/* stop the timer here... */
	}

	DBFEXIT;
	return 0;
}

int rfcomm_recv_cmd(rfcomm_sn *sn, struct sk_buff *skb)
{
	short_frame	*pkt = (short_frame*)skb->data;
	mcc_frame	*cmd;
	int		err = 0;

	DBFENTER;

	/* packet is correct here - was checked earlier */
	if (pkt->length.ea) {
		skb_pull(skb, SHORT_HDR_SIZE);
	} else {
		skb_pull(skb, LONG_HDR_SIZE);
	}
	cmd = (void*)skb->data;
	if (skb->len < sizeof(mcc_frame) && 
			skb->len < cmd->length.len + sizeof(mcc_frame)) {
		BTDEBUG("cmd packet too small, drop it...\n");
		kfree_skb(skb);
		return -EINVAL;
	}
	skb_pull(skb, sizeof(mcc_frame));
	skb_trim(skb, cmd->length.len);

	DBPRT(DBRECV, "RFCOMM %s: %x\n",(cmd->type.cr)?"cmd":"rsp",
			cmd->type.value);

	switch (cmd->type.value) {
		case R_MSC:
			recv_cmd_msc(sn, skb, cmd->type.cr);
			break;

		case RPN:
			recv_cmd_rpn(sn, skb, cmd->type.cr);
			break;

		case RLS:
			recv_cmd_rls(sn, skb, cmd->type.cr);
			break;

		case PN:
			recv_cmd_pn(sn, skb, cmd->type.cr);
			break;

		case FCON:
			recv_cmd_fc(sn, 1, cmd->type.cr);
			break;

		case FCOFF:
			recv_cmd_fc(sn, 0, cmd->type.cr);
			break;

		case NSC:
			/* check what was a command */
			break;

		default:
			DBPRT(DBRECV, "Unsupported command: %#02x\n", cmd->type.value);
			send_cmd_nsc(sn, cmd->type.value, cmd->type.cr);
	}

	kfree_skb(skb);
	DBFEXIT;
	return err;
}


/*
   these functions should send appropriate commands
   */
void rfcon_set_mcr(rfcomm_con *con, __u8 mcr)
{
	DBFENTER;

	if (con->modem.mr ^ mcr) {
		con->modem.mr = mcr;
		send_cmd_msc(con, MCC_CMD);
	}

	DBFEXIT;
}

void rfcon_set_flow(rfcomm_con *con, int flow)
{
	DBFENTER;
	con->flow = flow;
	if (con->modem.fc ^ !flow) {
		con->modem.fc = !flow;
		send_cmd_msc(con, MCC_CMD);
	}
	if (con->credit_based) {
		if (flow == FLOW_ON) {
			rfcon_control_ind(con, RFCOMM_BUFFER_REQUEST);
			if (con->local_credit < con->overall_credit) {
				rfcomm_send_credit(con, con->overall_credit - con->local_credit);
			}
		} else
			con->overall_credit = 0;
	}
	DBFEXIT;
}

void rfcon_update_flow(rfcomm_con *con)
{
	DBFENTER;
	if (con->credit_based) {
		if (con->flow == FLOW_ON) {
			if ((con->local_credit << 1) <= con->overall_credit) {
				rfcon_control_ind(con, RFCOMM_BUFFER_REQUEST);
				rfcomm_send_credit(con, con->overall_credit - con->local_credit);
			}
		}
	}
	DBFEXIT;
}

void rfcon_set_param(rfcomm_con *con, port_param *p)
{
	con->peer_param = *p;
}

void rfcon_set_break(rfcomm_con *con, __u8 b)
{
	con->break_signal.b1 = 1;
}


/* FCS table functions */

/* Calulates a reversed CRC table for the FCS check */

void create_crc_table(__u8 table[])
{
	int	i ,j;
	__u8 	data;
	__u8 	code_word = (__u8) 0xe0;  /*pol = x8+x2+x1+1*/
	__u8 	sr = (__u8) 0;            /*Shiftregister initiated to zero*/

	for (j=0; j < 256; j++) {
		data = (__u8) j;

		for (i=0;i<8;i++) {
			if((data & 0x1)^(sr & 0x1)) {
				sr >>= 1;
				sr ^= code_word;
			}
			else {
				sr >>= 1;
			}

			data >>= 1;
			sr &= 0xff;
		}

		table[j] = sr;
		sr = 0;
	} 
}

/* Calculates the checksum according to the RFCOMM specification */

__u8 crc_calc(__u8 *data, uint length)
{
	__u8	fcs = 0xff;

	DBFENTER;

	while (length--) {
		fcs=crctable[fcs^*data++];
	}

	/*Ones complement*/
	return 0xff-fcs;
}

#define CRC_VALID 0xcf

/* This functions check whether the checksum is correct or not. Length is
   the number of bytes in the message, data points to the beginning of the
   message */

__u32 crc_check(__u8 *data, uint length, __u8 check_sum)
{
	__u8	fcs = 0xff;

	while (length--) {
		fcs=crctable[fcs^*data++];
	}
	fcs=crctable[fcs^check_sum];
	if (fcs == (uint) 0xcf)/*CRC_VALID)*/{
		return 0;
	} else {
		DBPRT(DBRECV,"CRC INVALID: %#02x != 0xcf\n", fcs);
		return 1;
	}
}


/*
 * L2CAP client subsystem.
 * 
 */

int __rfcomm_data_ind(l2cap_ch *ch, struct sk_buff *skb)
{
	rfcomm_sn	*sn = (rfcomm_sn*)ch->priv;
	short_frame	*pkt;
	int		err = -EINVAL, len;
	__u8		control;

	DBFENTER;

	DBPARSERFCOMM(DBRECV, skb->data, skb->len, FROM_HOSTCTRL);
	DBDUMPCHAR(DBRECV, skb->data, skb->len);
	DBPRTPKT(DBRECV, skb->data, skb->len);
	
	/* process RFCOMM data packets */

	pkt = (short_frame*)skb->data;
	if (skb->len < SHORT_HDR_SIZE || (!pkt->length.ea && skb->len < LONG_HDR_SIZE)) {
		BTDEBUG("RFCOMM packet too short, drop it...\n");
		goto exit;
	}
		
	control = CLR_PF(pkt->control);

	if (pkt->length.ea) { /* short */
		len = pkt->length.len + SHORT_HDR_SIZE;
	} else {
		len = GET_LENGTH(((long_frame*)pkt)->length.len) + LONG_HDR_SIZE;
	}
	len += FCS_SIZE;

	if (control == UIH) {
		if (pkt->addr.dlci != 0)
			len += GET_PF(pkt->control); /* credit based flow control */
		if (skb->len < len) {
			BTDEBUG("RFCOMM packet too short, drop it...\n");
			goto exit;
		}
		DBPRT(DBRECV, "frame length: %d\n", len);
		err =  crc_check((__u8*)pkt, UIH_CRC_CHECK, skb->data[len-1]);
	} else
		err =  crc_check((__u8*)pkt, CTRL_CRC_CHECK, pkt->data[0]);

	if (err) {
		/* FCS error */
		BTDEBUG("RFCOMM pkt w/ wrong checksum. drop it...\n");
		goto exit;
	}

	/* 
	 * cut packet and remove FCS 
	 * each packet has only one command anyway
	 */
	skb_trim(skb, len-1);

	rfcomm_hold(sn);
	switch(CLR_PF(pkt->control)) {
		case SABM:
			recv_sabm(sn, skb);
			break;
		case UA:
			recv_ua(sn, skb);
			break;
		case DM:
			recv_dm(sn, skb);
			break;
		case DISC:
			recv_disc(sn, skb);
			break;
		case UIH:
			if (pkt->addr.dlci == 0) /* we have a command */
				rfcomm_recv_cmd(sn, skb);
			else
				rfcomm_recv_data(sn, skb);
			break;

		default:
			DBPRT(DBRECV, "Unknown frame format\n");
			rfcomm_put(sn);
			goto exit;
	}
	rfcomm_put(sn);
	DBFEXIT;
	return 0;
exit:
	kfree_skb(skb);
	DBFEXIT;
	return err;
}

int __rfcomm_connect_ind(l2cap_ch *ch)
{
	rfcomm_sn	*sn;

	DBFENTER;

	sn = rfcomm_create(&ch->bda);
	if (sn == NULL) {
		l2ca_connect_rsp(ch, CON_RSP_RESOURCE, 0);
		return -1;
	}
	sn->initiator = 0;
	sn->hdrlen = ch->hdrlen;
	l2ca_hold(ch);
	sn->ch = ch;
	ch->priv = sn;
	/*
	 * set configuration options, FIXME
	 */
	l2ca_setmtu(ch, LONG_HDR_SIZE + RFCOMM_GUESS_PAYLOAD_SIZE);
	l2ca_connect_rsp(ch, CON_RSP_SUCCESS, 0);

	ENTERSTATE(sn, CON_W4_CONREQ);
	rfcomm_start_timer(sn, RFCOMM_CONREQ_TIMEOUT);
	rfcomm_put(sn);

	DBFEXIT;
	return 0;
}

int __rfcomm_connect_cfm(l2cap_ch *ch, __u16 result, __u16 status)
{
	rfcomm_sn	*sn = (rfcomm_sn*)ch->priv;

	DBFENTER;
	rfcomm_hold(sn);

	if (result) {
		ENTERSTATE(sn, DEAD);
		rfcomm_connect_cfm(sn, STATUS_FAILURE);
		goto exit;
	}
	sn->hdrlen = ch->hdrlen;
	if (sn->initiator) {
		send_sabm(sn, 0);
		rfcomm_start_timer(sn, T1_TIMEOUT);	/* T1_TIMEOUT */
	}
exit:	
	rfcomm_put(sn);
	DBFEXIT;
	return 0;
}


int __rfcomm_config_ind(l2cap_ch *ch)
{
	rfcomm_sn	*sn = (rfcomm_sn*)ch->priv;
	__u16		result = 0;

	DBFENTER;

	if (sn == NULL)
		return -ENODEV;

	/* FIXME:
	 * set configuration options here for config_req
	 */
	result = l2ca_config_rsp(ch, CFG_RSP_SUCCESS);

	DBFEXIT;
	return result;
}

int __rfcomm_config_cfm(l2cap_ch *ch, __u16 result)
{
	rfcomm_sn	*sn = (rfcomm_sn*)ch->priv;

	DBFENTER;

	if (result) {
		ENTERSTATE(sn, DEAD);
		rfcomm_disconnect_ind(sn);
	}

	DBFEXIT;
	return 0;
}

int __rfcomm_disconnect_ind(l2cap_ch *ch)
{
	rfcomm_sn	*sn = (rfcomm_sn*)ch->priv;

	DBFENTER;

	rfcomm_hold(sn);
	
	ENTERSTATE(sn, DEAD);
	rfcomm_stop_timer(sn);
	rfcomm_disconnect_ind(sn);

	rfcomm_put(sn);
	
	DBFEXIT;
	return 0;
}

/* Callback functions for L2CAP layer */
l2cap_proto_ops rfcomm_ops = {
owner:		THIS_MODULE,
data_ind:	__rfcomm_data_ind,
connect_ind:	__rfcomm_connect_ind,
connect_cfm:	__rfcomm_connect_cfm,
config_ind:	__rfcomm_config_ind,
config_cfm:	__rfcomm_config_cfm,
disconnect_ind:	__rfcomm_disconnect_ind
};

int rfcomm_proc_read(char *buf, char **start, off_t offset, int len)
{
	int 		count = 0;
	rfcomm_sn	*sn;
	rfcomm_con	*con;

	DBFENTER;

	count += sprintf(buf+count, "Sessions (%d):\n", rfcomm_sns.len);
	btl_read_lock(&rfcomm_sns);
	btl_for_each (sn, rfcomm_sns) {
		count += sprintf(buf+count, "sn: bda: %s, refcnt: %d, state: %d,\n    credit_based: %s, local_credit: %d, remote_credit: %d\n", 
				bda2str(&sn->bda), atomic_read(&sn->refcnt), STATE(sn),
				sn->credit_based?"yes":"no", sn->local_credit, sn->remote_credit);
		btl_read_lock(&sn->cons);
		btl_for_each (con, sn->cons) {
			count += sprintf(buf+count, "    con: dlci: %d, refcnt: %d, state: %d,\n         credit_based: %s, local_credit: %d, remote_credit: %d\n",
					con->dlci, atomic_read(&con->refcnt), STATE(con),
					con->credit_based?"yes":"no", con->local_credit, con->remote_credit);
		}
		btl_read_unlock(&sn->cons);
	}
	btl_read_unlock(&rfcomm_sns);

	DBFEXIT;
	return count;
}


struct proc_dir_entry	*rfcomm_proc;

int rfcomm_init(void)
{
	int	err;

	DBFENTER;

	/* Register RFCOMM protocol */
	err =  l2ca_register_protocol(PSM_RFCOMM, &rfcomm_ops);
	if (err != 0) {
		BTERROR("Unable to register RFCOMM protocol\n");
		goto err1;
	}

	rfcomm_proc = create_proc_info_entry("rfcomm", 0, proc_affix, rfcomm_proc_read);
	if (rfcomm_proc == NULL) {
		BTERROR("Unable to register proc fs entry\n");
		goto err2;
	}
	create_crc_table(crctable);
	btl_head_init(&rfcomm_protos);
	btl_head_init(&rfcomm_sns);

	return 0;
err2:
	l2ca_unregister_protocol(PSM_RFCOMM);
err1:
	DBFEXIT;
	return err;

}

void rfcomm_exit(void)
{
	DBFENTER;

	remove_proc_entry("rfcomm", proc_affix);
	l2ca_unregister_protocol(PSM_RFCOMM);

	DBFEXIT;
}

/**********************************************************************************/

int init_rpf(void);
void exit_rpf(void);

int __init init_rfcomm(void)
{
	int	err = 0;

	err = rfcomm_init();
	if (err != 0)
		return err;

	err = bty_init();
	if (err != 0)
		goto err1;

	err = init_rpf();
	if (err != 0)
		goto err2;

	printk("Affix Bluetooth RFCOMM Protocol loaded (affix_rfcomm)\n");
	printk("Copyright (C) 2001, 2002 Nokia Corporation\n");
	printk("Written by Dmitry Kasatkin <dmitry.kasatkin@nokia.com>\n");

	return 0;
err2:
	bty_exit();
err1:	
	rfcomm_exit();
	return err;
}

void __exit exit_rfcomm(void)
{
	exit_rpf();
	bty_exit();
	rfcomm_exit();
}

#ifdef MODULE
/*if we are in the kernel we call it from hci.c */
module_init(init_rfcomm);
module_exit(exit_rfcomm);

MODULE_AUTHOR("Dmitry Kasatkin <dmitry.kasatkin@nokia.com>");
MODULE_DESCRIPTION("Affix RFCOMM module");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif
MODULE_PARM(bty_maxdev, "i"); /* For insmod */

#endif // MODULE

EXPORT_NO_SYMBOLS;

